# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import importlib
import sys
from typing import Any, Optional, Union

import torch
import torch.multiprocessing as mp
from tabulate import tabulate
from torch import Tensor
from torch.nn import Module
from torch.optim import Optimizer
from torch.utils.data import DataLoader, TensorDataset
from torch.utils.data.distributed import DistributedSampler
from tqdm import tqdm

import mlrun

from .callbacks import (
    Callback,
    HyperparametersKeys,
    MLRunLoggingCallback,
    TensorboardLoggingCallback,
)
from .callbacks_handler import CallbacksHandler
from .utils import PyTorchTypes


class PyTorchMLRunInterface:
    """
    An interface for enabling convenient MLRun features for the PyTorch framework, including training, evaluating and
    automatic logging.
    """

    # MLRun's context default name:
    DEFAULT_CONTEXT_NAME = "mlrun-pytorch"

    def __init__(self, model: Module, context: mlrun.MLClientCtx = None):
        """
        Initialize an interface for running training and evaluation on the given parameters.

        :param model:   The model to train / evaluate.
        :param context: MLRun context to use. If None, the context will be taken from 'mlrun.get_or_create_ctx()'.
        """
        # Set the context:
        self._context = (
            context
            if context is not None
            else mlrun.get_or_create_ctx(self.DEFAULT_CONTEXT_NAME)
        )

        # Store the model:
        self._model = model

        # Prepare methods parameters:
        self._training_set = None  # type: DataLoader
        self._loss_function = None  # type: Module
        self._optimizer = None  # type: Optimizer
        self._validation_set = None  # type: DataLoader
        self._metric_functions = None  # type: List[PyTorchTypes.MetricFunctionType]
        self._scheduler = None
        self._scheduler_step_frequency = None  # type: int
        self._epochs = None  # type: int
        self._training_iterations = None  # type: int
        self._validation_iterations = None  # type: int
        self._callbacks = []  # type: List[Callback]
        self._use_cuda = None  # type: bool
        self._use_horovod = None  # type: bool

        # Prepare inner attributes:
        self._hvd = None
        self._training_sampler = None  # type: DistributedSampler
        self._validation_sampler = None  # type: DistributedSampler
        self._callbacks_handler = None  # type: CallbacksHandler

    @property
    def model(self) -> Module:
        """
        Get the model stored in the interface.

        :return: The interface model.
        """
        return self._model

    @property
    def context(self) -> mlrun.MLClientCtx:
        """
        Get the interface MLRun context.

        :return: The interface MLRun context.
        """
        return self._context

    def train(
        self,
        training_set: DataLoader,
        loss_function: Module,
        optimizer: Optimizer,
        validation_set: DataLoader = None,
        metric_functions: Optional[list[PyTorchTypes.MetricFunctionType]] = None,
        scheduler=None,
        scheduler_step_frequency: Union[int, float, str] = "epoch",
        epochs: int = 1,
        training_iterations: Optional[int] = None,
        validation_iterations: Optional[int] = None,
        callbacks: Optional[list[Callback]] = None,
        use_cuda: bool = True,
        use_horovod: Optional[bool] = None,
    ):
        """
        Initiate a training process on this interface configuration.

        :param training_set:             A data loader for the training process.
        :param loss_function:            The loss function to use during training.
        :param optimizer:                The optimizer to use during the training.
        :param validation_set:           A data loader for the validation process.
        :param metric_functions:         The metrics to use on training and validation.
        :param scheduler:                Scheduler to use on the optimizer at the end of each epoch. The scheduler must
                                         have a 'step' method with no input.
        :param scheduler_step_frequency: The frequency in which to step the given scheduler. Can be equal to one of the
                                         strings 'epoch' (for at the end of every epoch) and 'batch' (for at the end of
                                         every batch), or an integer that specify per how many iterations to step or a
                                         float percentage (0.0 < x < 1.0) for per x / iterations to step. Default:
                                         'epoch'.
        :param epochs:                   Amount of epochs to perform. Default: a single epoch.
        :param training_iterations:      Amount of iterations (batches) to perform on each epoch's training. If 'None'
                                         the entire training set will be used.
        :param validation_iterations:    Amount of iterations (batches) to perform on each epoch's validation. If 'None'
                                         the entire validation set will be used.
        :param callbacks:                The callbacks to use on this run.
        :param use_cuda:                 Whether to use cuda. Only relevant if cuda is available. Default: True.
        :param use_horovod:              Whether to use horovod - a distributed training framework. Default: None,
                                         meaning it will be read from context if available and if not - False.
        """
        # Load the input:
        self._parse_and_store(
            training_set=training_set,
            loss_function=loss_function,
            optimizer=optimizer,
            validation_set=validation_set,
            metric_functions=metric_functions,
            scheduler=scheduler,
            scheduler_step_frequency=scheduler_step_frequency,
            epochs=epochs,
            training_iterations=training_iterations,
            validation_iterations=validation_iterations,
            callbacks=callbacks,
            use_cuda=use_cuda,
            use_horovod=use_horovod,
        )

        # Set up the inner attributes (initializing horovod and creating the callbacks handler):
        self._setup()

        # Beginning of run callbacks:
        self._callbacks_handler.on_run_begin()

        # Start the epochs:
        for epoch in range(self._epochs):
            # Beginning of a epoch callbacks:
            self._callbacks_handler.on_epoch_begin(epoch=epoch)
            print(
                f"Epoch {str(epoch + 1).rjust(len(str(self._epochs)))}/{self._epochs}:"
            )

            # Train:
            self._callbacks_handler.on_train_begin()
            self._train()
            if not self._callbacks_handler.on_train_end():
                break

            # Validate:
            if self._validation_set is not None:
                self._callbacks_handler.on_validation_begin()
                loss_value, metric_values = self._validate()
                # If horovod is used, wait for all ranks to calculate the loss and metrics averages:
                if self._use_horovod:
                    loss_value = self._metric_average(
                        rank_value=loss_value,
                        name=f"average_{self._get_metric_name(metric=self._loss_function)}",
                    )
                    metric_values = [
                        self._metric_average(
                            rank_value=metric_value,
                            name=f"average_{self._get_metric_name(metric=metric_function)}",
                        )
                        for metric_value, metric_function in zip(
                            metric_values, self._metric_functions
                        )
                    ]
                self._print_results(loss_value=loss_value, metric_values=metric_values)
                if not self._callbacks_handler.on_validation_end(
                    loss_value=loss_value, metric_values=metric_values
                ):
                    break

            # End of an epoch callbacks:
            if not self._callbacks_handler.on_epoch_end(epoch=epoch):
                break
            print()

        # End of run callbacks:
        self._callbacks_handler.on_run_end()

        # Clear the interface:
        self._clear()

    def evaluate(
        self,
        dataset: DataLoader,
        loss_function: Module = None,
        metric_functions: Optional[list[PyTorchTypes.MetricFunctionType]] = None,
        iterations: Optional[int] = None,
        callbacks: Optional[list[Callback]] = None,
        use_cuda: bool = True,
        use_horovod: Optional[bool] = None,
    ) -> list[PyTorchTypes.MetricValueType]:
        """
        Initiate an evaluation process on this interface configuration.

        :param dataset:          A data loader for the validation process.
        :param loss_function:    The loss function to use during training.
        :param metric_functions: The metrics to use on training and validation.
        :param iterations:       Amount of iterations (batches) to perform on the dataset. If 'None' the entire dataset
                                 will be used.
        :param callbacks:        The callbacks to use on this run.
        :param use_cuda:         Whether or not to use cuda. Only relevant if cuda is available. Default: True.
        :param use_horovod:      Whether or not to use horovod - a distributed training framework. Default: None,
                                 meaning it will be read from context if available and if not - False.

        :return: The evaluation loss and metrics results in a list.
        """
        # Load the input:
        self._parse_and_store(
            validation_set=dataset,
            loss_function=loss_function,
            metric_functions=metric_functions,
            validation_iterations=iterations,
            callbacks=callbacks,
            use_cuda=use_cuda,
            use_horovod=use_horovod,
        )

        # Setup the inner attributes (initializing horovod and creating the callbacks handler):
        self._setup()

        # Beginning of run callbacks:
        self._callbacks_handler.on_run_begin()

        # Beginning of a epoch callbacks (Only one epoch in evaluation):
        self._callbacks_handler.on_epoch_begin(epoch=1)

        # Evaluate:
        self._callbacks_handler.on_validation_begin()
        loss_value, metric_values = self._validate(is_evaluation=True)

        # If horovod is used, wait for all ranks to calculate the loss and metrics averages:
        if self._use_horovod:
            loss_value = self._metric_average(
                rank_value=loss_value,
                name=f"average_{self._get_metric_name(metric=self._loss_function)}",
            )
            metric_values = [
                self._metric_average(
                    rank_value=metric_value,
                    name=f"average_{self._get_metric_name(metric=metric_function)}",
                )
                for metric_value, metric_function in zip(
                    metric_values, self._metric_functions
                )
            ]

        # End the validation:
        self._print_results(loss_value=loss_value, metric_values=metric_values)
        self._callbacks_handler.on_validation_end(
            loss_value=loss_value, metric_values=metric_values
        )

        # End of a epoch callbacks:
        self._callbacks_handler.on_epoch_end(epoch=1)
        print()

        # End of run callbacks:
        self._callbacks_handler.on_run_end()

        # Clear the interface:
        self._clear()

        return [loss_value] + metric_values

    def add_auto_logging_callbacks(
        self,
        add_mlrun_logger: bool = True,
        mlrun_callback_kwargs: Optional[dict[str, Any]] = None,
        add_tensorboard_logger: bool = True,
        tensorboard_callback_kwargs: Optional[dict[str, Any]] = None,
    ):
        """
        Get automatic logging callbacks to both MLRun's context and Tensorboard. For further features of logging to both
        MLRun and Tensorboard, see 'pytorch.callbacks.MLRunLoggingCallback' and
        'pytorch.callbacks.TensorboardLoggingCallback'.

        :param add_mlrun_logger:            Whether or not to add the 'MLRunLoggingCallback'. Default: True.
        :param mlrun_callback_kwargs:       Key word arguments for the MLRun callback. For further information see the
                                            documentation of the class 'MLRunLoggingCallback'. Note that both 'context'
                                            and 'auto_log' parameters are already given here.
        :param add_tensorboard_logger:      Whether or not to add the 'TensorboardLoggingCallback'. Default: True.
        :param tensorboard_callback_kwargs: Key word arguments for the tensorboard callback. For further information see
                                            the documentation of the class 'TensorboardLoggingCallback'. Note that both
                                            'context' and 'auto_log' parameters are already given here.
        """
        # Set the dictionaries defaults:
        mlrun_callback_kwargs = (
            {} if mlrun_callback_kwargs is None else mlrun_callback_kwargs
        )
        tensorboard_callback_kwargs = (
            {} if tensorboard_callback_kwargs is None else tensorboard_callback_kwargs
        )

        # Add the loggers:
        if add_mlrun_logger:
            # Add the MLRun logging callback:
            self._callbacks.append(
                MLRunLoggingCallback(
                    context=self._context, auto_log=True, **mlrun_callback_kwargs
                )
            )
        if add_tensorboard_logger:
            # Add the Tensorboard logging callback:
            self._callbacks.append(
                TensorboardLoggingCallback(
                    context=self._context, auto_log=True, **tensorboard_callback_kwargs
                )
            )

    def predict(
        self,
        inputs: Union[Tensor, list[Tensor]],
        use_cuda: bool = True,
        batch_size: int = -1,
    ) -> Tensor:
        """
        Run prediction on the given data. Batched data can be predicted as well.

        :param inputs:     The inputs to infer through the model and get its predictions. Expecting a torch.Tensor or a
                           list of torch.Tensors to match each input layer.
        :param use_cuda:   Whether or not to use cuda. Only relevant if cuda is available. Default: True.
        :param batch_size: Batch size to use for prediction. If equals to -1, the entire inputs will be inferred at once
                           (batch size will be equal to the amount of inputs). Default: -1.

        :return: The model's predictions (outputs) list.
        """
        # Move the model to cuda if needed:
        if use_cuda and torch.cuda.is_available():
            self._objects_to_cuda()

        # Set model to evaluate mode:
        self._model.eval()

        # Wrap in a list if given as a single Tensor:
        if not isinstance(inputs, list):
            inputs = [inputs]

        # Initialize a data loader for the given inputs:
        data_loader = DataLoader(
            TensorDataset(*inputs),
            batch_size=batch_size if batch_size != -1 else len(inputs),
        )

        # Start the inference:
        with torch.no_grad():
            predictions = None  # type: Tensor
            for x in data_loader:
                # Move the input tensor to cuda if needed:
                if use_cuda and torch.cuda.is_available():
                    x = self._tensor_to_cuda(tensor=x)
                # Get the model's prediction:
                y = self._model(*x)
                # Store the predictions one by one:
                if predictions is None:
                    predictions = y
                else:
                    predictions = torch.cat((predictions, y))

        return predictions

    def _parse_and_store(
        self,
        training_set: DataLoader = None,
        loss_function: Module = None,
        optimizer: Optimizer = None,
        validation_set: DataLoader = None,
        metric_functions: Optional[list[PyTorchTypes.MetricFunctionType]] = None,
        scheduler=None,
        scheduler_step_frequency: Union[int, float, str] = "epoch",
        epochs: int = 1,
        training_iterations: Optional[int] = None,
        validation_iterations: Optional[int] = None,
        callbacks: Optional[list[Callback]] = None,
        use_cuda: bool = True,
        use_horovod: Optional[bool] = None,
    ):
        """
        Parse and store the given input so the interface can starting training / evaluating.

        :param training_set:             A data loader for the training process.
        :param loss_function:            The loss function to use during training.
        :param optimizer:                The optimizer to use during the training.
        :param validation_set:           A data loader for the validation process.
        :param metric_functions:         The metrics to use on training and validation.
        :param scheduler:                Scheduler to use on the optimizer at the end of each epoch. The scheduler must
                                         have a 'step' method with no input.
        :param scheduler_step_frequency: The frequecny in which to step the given scheduler. Can be equal to one of the
                                         strings 'epoch' (for at the end of every epoch) and 'batch' (for at the end of
                                         every batch), or an integer that specify per how many iterations to step or a
                                         float percentage (0.0 < x < 1.0) for per x / iterations to step. Default:
                                         'epoch'.
        :param epochs:                   Amount of epochs to perform. Default: a single epoch.
        :param training_iterations:      Amount of iterations (batches) to perform on each epoch's training. If 'None'
                                         the entire training set will be used.
        :param validation_iterations:    Amount of iterations (batches) to perform on each epoch's validation. If 'None'
                                         the entire validation set will be used.
        :param callbacks:                The callbacks to use on this run.
        :param use_cuda:                 Whether or not to use cuda. Only relevant if cuda is available. Default:
                                         True.
        :param use_horovod:              Whether or not to use horovod - a distributed training framework. Default:
                                         None, meaning it will be read from context if available and if not - False.

        :raise MLRunInvalidArgumentError: In case one of the given parameters is invalid.
        """
        # Parse and validate input:
        # # Metric functions:
        if metric_functions is None:
            metric_functions = []
        # # Training iterations:
        if training_set is not None:
            if training_iterations is None:
                training_iterations = len(training_set)
            elif training_iterations < 1:
                raise mlrun.errors.MLRunInvalidArgumentError(
                    f"The 'training_iterations' parameter must be bigger or equal to one, received: "
                    f"{training_iterations}"
                )
            elif training_iterations > len(training_set):
                raise mlrun.errors.MLRunInvalidArgumentError(
                    f"The 'training_iterations' cannot be bigger than the given training dataset. The size of "
                    f"the given training set is {len(training_set)} yet the received iterations parameter is "
                    f"{training_iterations}."
                )
        # # Validation iterations:
        if validation_set is not None:
            if validation_iterations is None:
                validation_iterations = len(validation_set)
            elif validation_iterations < 1:
                raise mlrun.errors.MLRunInvalidArgumentError(
                    f"The 'validation_iterations' parameter must be bigger or equal to one, "
                    f"received: {validation_iterations}"
                )
            elif validation_iterations > len(validation_set):
                raise mlrun.errors.MLRunInvalidArgumentError(
                    f"The 'validation_iterations' cannot be bigger than the given validation dataset. The size of the "
                    f"given validation set is {len(validation_set)} yet the received iterations parameter is "
                    f"{validation_iterations}."
                )
        # # Epochs:
        if epochs < 1:
            raise mlrun.errors.MLRunInvalidArgumentError(
                f"The 'epochs' parameter must be bigger or equal to one, received: {epochs}"
            )
        # # Scheduler step frequency:
        if isinstance(scheduler_step_frequency, str):
            if scheduler_step_frequency == "epoch":
                scheduler_step_frequency = training_iterations
            elif scheduler_step_frequency == "batch":
                scheduler_step_frequency = 1
            else:
                raise mlrun.errors.MLRunInvalidArgumentError(
                    f"The scheduler step frequency parameter can be passed as a string of two values: "
                    f"'epoch' or 'batch', but the value given was: '{scheduler_step_frequency}'"
                )
        elif isinstance(scheduler_step_frequency, float):
            if scheduler_step_frequency < 0.0 or scheduler_step_frequency > 1.0:
                raise mlrun.errors.MLRunInvalidArgumentError(
                    f"The scheduler step frequency parameter can be passed as a float with value between "
                    f"0.0 to 1.0, but the value given was: '{scheduler_step_frequency}'"
                )
            scheduler_step_frequency = int(
                training_iterations * scheduler_step_frequency
            )
        # # Callbacks:
        if callbacks is None:
            callbacks = []
        # # Use horovod:
        if use_horovod is None:
            use_horovod = (
                self._context.labels.get("kind", "") == "mpijob"
                if self._context is not None
                else False
            )

        # Store the configurations:
        self._training_set = training_set
        self._loss_function = loss_function
        self._optimizer = optimizer
        self._validation_set = validation_set
        self._metric_functions = metric_functions
        self._scheduler = scheduler
        self._scheduler_step_frequency = scheduler_step_frequency
        self._epochs = epochs
        self._training_iterations = training_iterations
        self._validation_iterations = validation_iterations
        self._callbacks += callbacks
        self._use_cuda = use_cuda
        self._use_horovod = use_horovod

    def _objects_to_cuda(self):
        """
        Copy the interface objects - model, loss, optimizer and scheduler to cuda memory.
        """
        # Model:
        if not self._is_module_in_cuda(module=self._model):
            self._model = self._model.cuda()

        # Loss:
        if self._loss_function is not None and not self._is_module_in_cuda(
            module=self._loss_function
        ):
            self._loss_function = self._loss_function.cuda()

        # Optimizer:
        if self._optimizer is not None:
            for key in self._optimizer.state.keys():
                self._optimizer.state[key] = self._tensor_to_cuda(
                    tensor=self._optimizer.state[key]
                )

        # Scheduler:
        if self._scheduler is not None:
            for key in self._scheduler.__dict__.keys():
                self._scheduler.__dict__[key] = self._tensor_to_cuda(
                    tensor=self._scheduler.__dict__[key]
                )

    def _setup(self):
        """
        Setup the inner attributes of the interface, initializing horovod and the callbacks handler. This method must be
        called before train and evaluate.
        """
        # Setup horovod:
        if self._use_horovod:
            # Import horovod:
            self._hvd = importlib.import_module("horovod.torch")
            # Initialize horovod:
            self._hvd.init()
            # Limit the number of CPU threads to be used per worker:
            torch.set_num_threads(1)

        # Setup additional multiprocessing related key word arguments for the data loaders initialization:
        mp_data_loader_kwargs = {}

        # Setup cuda:
        if self._use_cuda and torch.cuda.is_available():
            if self._use_horovod:
                # Set the torch environment to use a specific GPU according to the horovod worker's local rank:
                torch.cuda.set_device(self._hvd.local_rank())
                # Log horovod worker device:
                print(
                    f"Horovod worker #{self._hvd.rank()} is using GPU:{self._hvd.local_rank()}"
                )
                # Register the required multiprocessing arguments:
                mp_data_loader_kwargs["num_workers"] = 1
                mp_data_loader_kwargs["pin_memory"] = True
            # Move the model and the stored objects to the GPU:
            self._objects_to_cuda()
        elif self._use_horovod:
            # Log horovod worker device:
            print(f"Horovod worker #{self._hvd.rank()} is using CPU")

        # Initialize a callbacks handler:
        if self._use_horovod:
            self._callbacks_handler = CallbacksHandler(
                callbacks=[
                    callback
                    for callback in self._callbacks
                    if callback.on_horovod_check(rank=self._hvd.rank())
                ]
            )
        else:
            self._callbacks_handler = CallbacksHandler(callbacks=self._callbacks)

        # Prepare horovod for the run if needed:
        if self._use_horovod:
            # When supported, use 'forkserver' to spawn dataloader workers instead of 'fork' to prevent issues with
            # Infiniband implementations that are not fork-safe:
            if (
                mp_data_loader_kwargs.get("num_workers", 0) > 0
                and hasattr(mp, "_supports_context")
                and mp._supports_context
                and "forkserver" in mp.get_all_start_methods()
            ):
                mp_data_loader_kwargs["multiprocessing_context"] = "forkserver"
            # Partition dataset among workers using distributed samplers:
            if self._training_set is not None:
                self._training_sampler = DistributedSampler(
                    self._training_set.dataset,
                    num_replicas=self._hvd.size(),
                    rank=self._hvd.rank(),
                )
                self._training_set = self._insert_sampler_to_data_loader(
                    data_loader=self._training_set,
                    sampler=self._training_sampler,
                    multiprocessing_kwargs=mp_data_loader_kwargs,
                )
            if self._validation_set is not None:
                self._validation_sampler = DistributedSampler(
                    self._validation_set.dataset,
                    num_replicas=self._hvd.size(),
                    rank=self._hvd.rank(),
                )
                self._validation_set = self._insert_sampler_to_data_loader(
                    data_loader=self._validation_set,
                    sampler=self._validation_sampler,
                    multiprocessing_kwargs=mp_data_loader_kwargs,
                )
            # Broadcast parameters and optimizer state:
            self._hvd.broadcast_parameters(self._model.state_dict(), root_rank=0)
            if self._optimizer is not None:
                self._hvd.broadcast_optimizer_state(self._optimizer, root_rank=0)
                # Add Horovod Distributed Optimizer:
                self._optimizer = self._hvd.DistributedOptimizer(
                    self._optimizer, named_parameters=self._model.named_parameters()
                )

        # Setup the callbacks functions:
        self._callbacks_handler.on_setup(
            model=self._model,
            training_set=self._training_set,
            validation_set=self._validation_set,
            loss_function=self._loss_function,
            optimizer=self._optimizer,
            metric_functions=self._metric_functions,
            scheduler=self._scheduler,
        )

    def _train(self):
        """
        Initiate a single epoch training.
        """
        # Set model to train mode:
        self._model.train()

        # Start the training:
        progress_bar = self._create_progress_bar(
            dataset=self._training_set,
            iterations=self._training_iterations,
            description="Training",
            metrics=[self._loss_function] + self._metric_functions,
        )
        for batch, (x, y_true) in progress_bar:
            # Check if iteration exceeded:
            if batch == self._training_iterations:
                break

            # Move to GPU if needed:
            if self._use_cuda and torch.cuda.is_available():
                x, y_true = self._tensor_to_cuda(tensor=(x, y_true))

            # Beginning of a batch callbacks:
            self._callbacks_handler.on_train_batch_begin(
                batch=batch, x=x, y_true=y_true
            )

            # Zero the parameters gradients:
            self._optimizer.zero_grad()

            # Infer the input:
            self._callbacks_handler.on_inference_begin(x=x)
            y_pred = self._model(x)
            self._callbacks_handler.on_inference_end(y_pred=y_pred, y_true=y_true)

            # Calculate loss:
            self._callbacks_handler.on_train_loss_begin()
            loss_value = self._loss_function(y_pred, y_true)
            self._callbacks_handler.on_train_loss_end(loss_value=loss_value)

            # Measure accuracies:
            self._callbacks_handler.on_train_metrics_begin()
            metric_values = self._metrics(y_pred=y_pred, y_true=y_true)
            self._callbacks_handler.on_train_metrics_end(metric_values=metric_values)

            # Update the progress bar with the recent values:
            self._update_progress_bar(
                progress_bar=progress_bar,
                metrics=[self._loss_function] + self._metric_functions,
                values=[loss_value] + metric_values,
            )

            # Perform backward propagation:
            self._callbacks_handler.on_backward_begin()
            loss_value.backward()
            self._callbacks_handler.on_backward_end()

            # Step optimizer:
            self._callbacks_handler.on_optimizer_step_begin()
            self._optimizer.step()
            self._callbacks_handler.on_optimizer_step_end()

            # Step scheduler:
            if (
                self._scheduler is not None
                and (batch + 1) % self._scheduler_step_frequency == 0
            ):
                self._callbacks_handler.on_scheduler_step_begin()
                self._scheduler.step()
                self._callbacks_handler.on_scheduler_step_end()

            # End of batch callbacks:
            if not self._callbacks_handler.on_train_batch_end(
                batch=batch, x=x, y_pred=y_pred, y_true=y_true
            ):
                break

    def _validate(
        self, is_evaluation: bool = False
    ) -> tuple[PyTorchTypes.MetricValueType, list[PyTorchTypes.MetricValueType]]:
        """
        Initiate a single epoch validation.

        :param is_evaluation: Whether or not this call is part of an evaluation or training.

        :return: A tuple of the validation summary:
                 [0] = Validation loss value summary.
                 [1] = A list of metrics summaries.
        """
        # Set model to evaluate mode:
        self._model.eval()

        # Start the validation:
        losses = []
        metrics = []
        progress_bar = self._create_progress_bar(
            dataset=self._validation_set,
            iterations=self._validation_iterations,
            description="Evaluating" if is_evaluation else "Validating",
            metrics=[self._loss_function] + self._metric_functions,
        )
        with torch.no_grad():
            for batch, (x, y_true) in progress_bar:
                # Check if iteration exceeded:
                if batch == self._validation_iterations:
                    break

                # Move to GPU if needed:
                if self._use_cuda and torch.cuda.is_available():
                    x, y_true = self._tensor_to_cuda(tensor=(x, y_true))

                # Beginning of a batch callbacks:
                self._callbacks_handler.on_validation_batch_begin(
                    batch=batch, x=x, y_true=y_true
                )

                # Infer the input:
                self._callbacks_handler.on_inference_begin(x=x)
                y_pred = self._model(x)
                self._callbacks_handler.on_inference_end(y_pred=y_pred, y_true=y_true)

                # Calculate loss:
                self._callbacks_handler.on_validation_loss_begin()
                loss_value = self._loss_function(y_pred, y_true)
                self._callbacks_handler.on_validation_loss_end(loss_value=loss_value)

                # Measure accuracies:
                self._callbacks_handler.on_validation_metrics_begin()
                metric_values = self._metrics(y_pred=y_pred, y_true=y_true)
                self._callbacks_handler.on_validation_metrics_end(
                    metric_values=metric_values
                )

                # Update the progress bar with the recent values:
                self._update_progress_bar(
                    progress_bar=progress_bar,
                    metrics=[self._loss_function] + self._metric_functions,
                    values=[loss_value] + metric_values,
                )

                # Collect results:
                losses.append(loss_value)
                metrics.append(metric_values)

                # End of batch callbacks:
                if not self._callbacks_handler.on_validation_batch_end(
                    batch=batch,
                    x=x,
                    y_pred=y_pred,
                    y_true=y_true,
                ):
                    break

        # Calculate the final average of the loss and accuracy values:
        loss_value = sum(losses) / len(losses)
        metric_values = (
            [(sum(metric) / len(metric)) for metric in metrics]
            if len(metrics) > 0
            else []
        )
        return loss_value, metric_values

    def _print_results(self, loss_value: Tensor, metric_values: list[float]):
        """
        Print the given result between each epoch.

        :param loss_value:    The loss result to print.
        :param metric_values: The metrics result to print.
        """
        table = [[self._get_metric_name(metric=self._loss_function), float(loss_value)]]
        for metric_function, metric_value in zip(self._metric_functions, metric_values):
            table.append([self._get_metric_name(metric=metric_function), metric_value])
        print(
            "\nSummary:\n"
            + tabulate(table, headers=["Metrics", "Values"], tablefmt="pretty")
        )

    def _metrics(self, y_pred: Tensor, y_true: Tensor) -> list[float]:
        """
        Call all the metrics on the given batch's truth and prediction output.

        :param y_pred: The batch's truth value.
        :param y_true: The model's output for the related input.

        :return: A list with each metric result.
        """
        accuracies = []
        for metric_function in self._metric_functions:
            accuracies.append(metric_function(y_pred, y_true))
        return accuracies

    def _metric_average(self, rank_value: Union[Tensor, float], name: str) -> float:
        """
        Wait for all ranks and calculate the average of the metric provided.

        :param rank_value: The caller rank metric value.
        :param name:       The metric name.

        :return: The metric average across all ranks.
        """
        if not isinstance(rank_value, Tensor):
            rank_value = torch.tensor(rank_value)
        average_tensor = self._hvd.allreduce(rank_value, name=name)
        return average_tensor.item()

    def _get_learning_rate(self) -> Union[tuple[str, list[Union[str, int]]], None]:
        """
        Try and get the learning rate value form the stored optimizer.

        :return: The key chain to get the optimizer learning rate value or None if the learning rate could not be
                 accessed via the common key.
        """
        if "lr" in self._optimizer.param_groups[0]:
            return HyperparametersKeys.OPTIMIZER, ["param_groups", 0, "lr"]
        if "learning_rate" in self._optimizer.param_groups[0]:
            return HyperparametersKeys.OPTIMIZER, ["param_groups", 0, "learning_rate"]
        return None

    def _clear(self):
        """
        Clear the interface from the methods parameters, setting them back to None. The interface's model will not be
        touched.
        """
        # Clear the methods parameters:
        self._training_set = None  # type: DataLoader
        self._loss_function = None  # type: Module
        self._optimizer = None  # type: Optimizer
        self._validation_set = None  # type: DataLoader
        self._metric_functions = None  # type: List[PyTorchTypes.MetricFunctionType]
        self._scheduler = None
        self._scheduler_step_frequency = None  # type: int
        self._epochs = None  # type: int
        self._training_iterations = None  # type: int
        self._validation_iterations = None  # type: int
        self._callbacks = []  # type: List[Callback]
        self._use_cuda = None  # type: bool
        self._use_horovod = None  # type: bool

        # Clear the inner attributes:
        self._hvd = None
        self._training_sampler = None  # type: DistributedSampler
        self._validation_sampler = None  # type: DistributedSampler
        self._callbacks_handler = None  # type: CallbacksHandler

    @staticmethod
    def _insert_sampler_to_data_loader(
        data_loader: DataLoader,
        sampler: DistributedSampler,
        multiprocessing_kwargs: dict,
    ):
        """
        Initialize a new data loader based on the given data loader with the given sampler.

        :param data_loader:            The data loader to insert the sampler to.
        :param sampler:                Sampler to insert.
        :param multiprocessing_kwargs: Additional keyword arguments for the multiprocessing attributes.

        :return: The sampler initialized data loader.
        """
        pin_memory = multiprocessing_kwargs.get("pin_memory", data_loader.pin_memory)
        num_workers = multiprocessing_kwargs.get("num_workers", data_loader.num_workers)
        multiprocessing_context = multiprocessing_kwargs.get(
            "multiprocessing_context", data_loader.multiprocessing_context
        )
        with_sampler_data_loader = DataLoader(
            dataset=data_loader.dataset,
            batch_size=data_loader.batch_size,
            shuffle=False,
            sampler=sampler,
            num_workers=num_workers,
            collate_fn=data_loader.collate_fn,
            pin_memory=pin_memory,
            drop_last=data_loader.drop_last,
            timeout=data_loader.timeout,
            worker_init_fn=data_loader.worker_init_fn,
            multiprocessing_context=multiprocessing_context,
            generator=data_loader.generator,
            prefetch_factor=data_loader.prefetch_factor,
            persistent_workers=data_loader.persistent_workers,
        )
        del data_loader
        return with_sampler_data_loader

    @staticmethod
    def _is_module_in_cuda(module: Module) -> bool:
        """
        Check whether or not the module is in CUDA memory.

        :return: True if the module is in CUDA memory and False otherwise.
        """
        return next(module.parameters()).is_cuda

    @staticmethod
    def _tensor_to_cuda(
        tensor: Union[Tensor, dict, list, tuple],
    ) -> Union[Tensor, dict, list, tuple]:
        """
        Send to given tensor to cuda if it is a tensor. If the given object is a dictionary, the dictionary values will
        be sent to the function again recursively. If the given object is a list or a tuple, all the values in it will
        be sent as well. If the given object is not of type torch.Tensor at the end, nothing will happen.

        :param tensor: The batch to sent to cuda.

        :return: The copied tensor in cuda memory.
        """
        if isinstance(tensor, Tensor) and not tensor.is_cuda:
            tensor = tensor.cuda()
            if tensor._grad is not None:
                tensor._grad.data = tensor._grad.data.cuda()
        elif isinstance(tensor, dict):
            for key in tensor:
                tensor[key] = PyTorchMLRunInterface._tensor_to_cuda(tensor=tensor[key])
        elif isinstance(tensor, list):
            for index in range(len(tensor)):
                tensor[index] = PyTorchMLRunInterface._tensor_to_cuda(
                    tensor=tensor[index]
                )
        elif isinstance(tensor, tuple):
            cuda_tensor = ()
            for value in tensor:
                cuda_tensor += (PyTorchMLRunInterface._tensor_to_cuda(tensor=value),)
            tensor = cuda_tensor
        return tensor

    @staticmethod
    def _get_metric_name(metric: PyTorchTypes.MetricFunctionType) -> str:
        """
        Get the given metric function name.

        :param metric: The metric function pointer.

        :return: The metric name.
        """
        if isinstance(metric, Module):
            return metric.__class__.__name__
        return metric.__name__

    @staticmethod
    def _create_progress_bar(
        dataset: DataLoader,
        iterations: int,
        description: str,
        metrics: list[PyTorchTypes.MetricFunctionType],
    ) -> tqdm:
        """
        Create a progress bar for training and validating / evaluating.

        :param dataset:     The dataset to enumerate in this progress bar.
        :param iterations:  The amount of iterations that will be performed.
        :param description: The header appearing in the left most side of the progress bar.
        :param metrics:     Metrics to note in the right most side of the progress bar.

        :return: The created progress bar.
        """
        return tqdm(
            iterable=enumerate(dataset),
            bar_format="{desc}: {percentage:3.0f}%"
            " |{bar}| {n_fmt}/{total_fmt} "
            "[{elapsed}<{remaining}, {rate_fmt}{postfix}]",
            desc=description,
            postfix={
                PyTorchMLRunInterface._get_metric_name(metric=metric): "?"
                for metric in metrics
            },
            unit="Batch",
            total=iterations,
            ascii=False,
            file=sys.stdout,
        )

    @staticmethod
    def _update_progress_bar(
        progress_bar: tqdm,
        metrics: list[PyTorchTypes.MetricFunctionType],
        values: list[PyTorchTypes.MetricValueType],
    ):
        """
        Update the progress bar metrics results.

        :param progress_bar: The progress bar to update.
        :param metrics:      The metrics list noted in the progress bar.
        :param values:       The metrics recent calculated values.
        """
        progress_bar.set_postfix(
            ordered_dict={
                PyTorchMLRunInterface._get_metric_name(metric=metric): float(value)
                for metric, value in zip(metrics, values)
            },
            refresh=False,
        )



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import pathlib
import tempfile
from abc import ABC, abstractmethod
from typing import Any, Optional, Union

import numpy as np
import pandas as pd

from mlrun.artifacts import Artifact, DatasetArtifact
from mlrun.datastore import DataItem
from mlrun.errors import MLRunInvalidArgumentError

from ..utils import ArtifactType, SupportedFormat
from .default_packager import DefaultPackager

# Type for collection of numpy arrays (list / dict of arrays):
NumPyArrayCollectionType = Union[list[np.ndarray], dict[str, np.ndarray]]


class _Formatter(ABC):
    """
    An abstract class for a numpy formatter - supporting saving and loading arrays to and from specific file type.
    """

    @classmethod
    @abstractmethod
    def save(
        cls,
        obj: Union[np.ndarray, NumPyArrayCollectionType],
        file_path: str,
        **save_kwargs: dict,
    ):
        """
        Save the given array to the file path given.

        :param obj:         The numpy array to save.
        :param file_path:   The file to save to.
        :param save_kwargs: Additional keyword arguments to pass to the relevant save function of numpy.
        """
        pass

    @classmethod
    @abstractmethod
    def load(
        cls, file_path: str, **load_kwargs: dict
    ) -> Union[np.ndarray, NumPyArrayCollectionType]:
        """
        Load the array from the given file path.

        :param file_path:   The file to load the array from.
        :param load_kwargs: Additional keyword arguments to pass to the relevant load function of numpy.

        :return: The loaded array.
        """
        pass


class _TXTFormatter(_Formatter):
    """
    A static class for managing numpy txt files.
    """

    @classmethod
    def save(cls, obj: np.ndarray, file_path: str, **save_kwargs: dict):
        """
        Save the given array to the file path given.

        :param obj:         The numpy array to save.
        :param file_path:   The file to save to.
        :param save_kwargs: Additional keyword arguments to pass to the relevant save function of numpy.

        :raise MLRunInvalidArgumentError: If the array is above 2D.
        """
        if len(obj.shape) > 2:
            raise MLRunInvalidArgumentError(
                f"Cannot save the given array to file. Only 1D and 2D arrays can be saved to text files but the given "
                f"array is {len(obj.shape)}D (shape of {obj.shape}). Please use 'npy' format instead."
            )
        np.savetxt(file_path, obj, **save_kwargs)

    @classmethod
    def load(cls, file_path: str, **load_kwargs: dict) -> np.ndarray:
        """
        Load the array from the given 'txt' file path.

        :param file_path:   The file to load the array from.
        :param load_kwargs: Additional keyword arguments to pass to the relevant load function of numpy.

        :return: The loaded array.
        """
        return np.loadtxt(file_path, **load_kwargs)


class _CSVFormatter(_TXTFormatter):
    """
    A static class for managing numpy csv files.
    """

    @classmethod
    def save(cls, obj: np.ndarray, file_path: str, **save_kwargs: dict):
        """
        Save the given array to the file path given.

        :param obj:         The numpy array to save.
        :param file_path:   The file to save to.
        :param save_kwargs: Additional keyword arguments to pass to the relevant save function of numpy.

        :raise MLRunInvalidArgumentError: If the array is above 2D.
        """
        super().save(obj=obj, file_path=file_path, **{"delimiter": ",", **save_kwargs})

    @classmethod
    def load(cls, file_path: str, **load_kwargs: dict) -> np.ndarray:
        """
        Load the array from the given 'txt' file path.

        :param file_path:   The file to load the array from.
        :param load_kwargs: Additional keyword arguments to pass to the relevant load function of numpy.

        :return: The loaded array.
        """
        return super().load(file_path=file_path, **{"delimiter": ",", **load_kwargs})


class _NPYFormatter(_Formatter):
    """
    A static class for managing numpy npy files.
    """

    @classmethod
    def save(cls, obj: np.ndarray, file_path: str, **save_kwargs: dict):
        """
        Save the given array to the file path given.

        :param obj:         The numpy array to save.
        :param file_path:   The file to save to.
        :param save_kwargs: Additional keyword arguments to pass to the relevant save function of numpy.
        """
        np.save(file_path, obj, **save_kwargs)

    @classmethod
    def load(cls, file_path: str, **load_kwargs: dict) -> np.ndarray:
        """
        Load the array from the given 'npy' file path.

        :param file_path:   The file to load the array from.
        :param load_kwargs: Additional keyword arguments to pass to the relevant load function of numpy.

        :return: The loaded array.
        """
        return np.load(file_path, **load_kwargs)


class _NPZFormatter(_Formatter):
    """
    A static class for managing numpy npz files.
    """

    @classmethod
    def save(
        cls,
        obj: NumPyArrayCollectionType,
        file_path: str,
        is_compressed: bool = False,
        **save_kwargs: dict,
    ):
        """
        Save the given array to the file path given.

        :param obj:           The numpy array to save.
        :param file_path:     The file to save to.
        :param is_compressed: Whether to save it as a compressed npz file.
        :param save_kwargs:   Additional keyword arguments to pass to the relevant save function of numpy.
        """
        save_function = np.savez_compressed if is_compressed else np.savez
        if isinstance(obj, list):
            save_function(file_path, *obj)
        else:
            save_function(file_path, **obj)

    @classmethod
    def load(cls, file_path: str, **load_kwargs: dict) -> dict[str, np.ndarray]:
        """
        Load the arrays from the given 'npz' file path.

        :param file_path:   The file to load the array from.
        :param load_kwargs: Additional keyword arguments to pass to the relevant load function of numpy.

        :return: The loaded arrays as a mapping (dictionary) of type `np.lib.npyio.NpzFile`.
        """
        return np.load(file_path, **load_kwargs)


class NumPySupportedFormat(SupportedFormat[_Formatter]):
    """
    Library of numpy formats (file extensions) supported by the NumPy packagers.
    """

    NPY = "npy"
    NPZ = "npz"
    TXT = "txt"
    GZ = "gz"
    CSV = "csv"

    _FORMAT_HANDLERS_MAP = {
        NPY: _NPYFormatter,
        NPZ: _NPZFormatter,
        TXT: _TXTFormatter,
        GZ: _TXTFormatter,  # 'gz' format handled the same as 'txt'.
        CSV: _CSVFormatter,
    }

    @classmethod
    def get_single_array_formats(cls) -> list[str]:
        """
        Get the supported formats for saving one numpy array.

        :return: A list of all the supported formats for saving one numpy array.
        """
        return [cls.NPY, cls.TXT, cls.GZ, cls.CSV]

    @classmethod
    def get_multi_array_formats(cls) -> list[str]:
        """
        Get the supported formats for saving a collection (multiple) numpy arrays - e.g. list of arrays or dictionary of
        arrays.

        :return: A list of all the supported formats for saving multiple numpy arrays.
        """
        return [cls.NPZ]


# Default file formats for numpy arrays file artifacts:
DEFAULT_NUMPY_ARRAY_FORMAT = NumPySupportedFormat.NPY
DEFAULT_NUMPPY_ARRAY_COLLECTION_FORMAT = NumPySupportedFormat.NPZ


class NumPyNDArrayPackager(DefaultPackager):
    """
    ``numpy.ndarray`` packager.
    """

    PACKABLE_OBJECT_TYPE = np.ndarray

    # The size of an array to be stored as a result, rather than a file in the `get_default_packing_artifact_type`
    # method:
    _ARRAY_SIZE_AS_RESULT = 10

    def get_default_packing_artifact_type(self, obj: np.ndarray) -> str:
        """
        Get the default artifact type. Will be a result if the array size is less than 10, otherwise file.

        :param obj: The about to be packed array.

        :return: The default artifact type.
        """
        if obj.size < self._ARRAY_SIZE_AS_RESULT:
            return ArtifactType.RESULT
        return ArtifactType.FILE

    def get_default_unpacking_artifact_type(self, data_item: DataItem) -> str:
        """
        Get the default artifact type used for unpacking. Returns dataset if the data item represents a
        `DatasetArtifact` and otherwise, file.

        :param data_item: The about to be unpacked data item.

        :return: The default artifact type.
        """
        is_artifact = data_item.get_artifact_type()
        if is_artifact and is_artifact == "datasets":
            return ArtifactType.DATASET
        return ArtifactType.FILE

    def pack_result(self, obj: np.ndarray, key: str) -> dict:
        """
        Pack an array as a result.

        :param obj: The array to pack and log.
        :param key: The result's key.

        :return: The result dictionary.
        """
        # If the array is a number (size of 1), then we'll lok it as a single number. Otherwise, log as a list result:
        if obj.size == 1:
            obj = obj.item()
        else:
            obj = obj.tolist()

        return super().pack_result(obj=obj, key=key)

    def pack_file(
        self,
        obj: np.ndarray,
        key: str,
        file_format: str = DEFAULT_NUMPY_ARRAY_FORMAT,
        **save_kwargs,
    ) -> tuple[Artifact, dict]:
        """
        Pack an array as a file by the given format.

        :param obj:         The aray to pack.
        :param key:         The key to use for the artifact.
        :param file_format: The file format to save as. Default is npy.
        :param save_kwargs: Additional keyword arguments to pass to the numpy save functions.

        :return: The packed artifact and instructions.
        """
        # Save to file:
        formatter = NumPySupportedFormat.get_format_handler(fmt=file_format)
        temp_directory = pathlib.Path(tempfile.mkdtemp())
        self.add_future_clearing_path(path=temp_directory)
        file_path = temp_directory / f"{key}.{file_format}"
        formatter.save(obj=obj, file_path=str(file_path), **save_kwargs)

        # Create the artifact and instructions (Note: only 'npy' format support saving object arrays and that will
        # require pickling, hence we set the required instruction):
        artifact = Artifact(key=key, src_path=os.path.abspath(file_path))
        instructions = {"file_format": file_format}
        if file_format == NumPySupportedFormat.NPY and obj.dtype == np.object_:
            instructions["allow_pickle"] = True

        return artifact, instructions

    def pack_dataset(
        self,
        obj: np.ndarray,
        key: str,
        file_format: str = "",
    ) -> tuple[Artifact, dict]:
        """
        Pack an array as a dataset.

        :param obj:         The aray to pack.
        :param key:         The key to use for the artifact.
        :param file_format: The file format to save as. Default is parquet.

        :return: The packed artifact and instructions.

        :raise MLRunInvalidArgumentError: IF the shape of the array is not 1D / 2D.
        """
        # Validate it's a 2D array:
        if len(obj.shape) > 2:
            raise MLRunInvalidArgumentError(
                f"Cannot log the given numpy array as a dataset. Only 2D arrays can be saved as dataset, but the array "
                f"is {len(obj.shape)}D (shape of {obj.shape}). Please specify to log it as a 'file' instead ('npy' "
                f"format) or as an 'object' (pickle)."
            )

        # Cast to a `pd.DataFrame`:
        data_frame = pd.DataFrame(data=obj)

        # Create the artifact:
        artifact = DatasetArtifact(key=key, df=data_frame, format=file_format)

        return artifact, {}

    def unpack_file(
        self,
        data_item: DataItem,
        file_format: Optional[str] = None,
        allow_pickle: bool = False,
    ) -> np.ndarray:
        """
        Unpack a numppy array from file.

        :param data_item:    The data item to unpack.
        :param file_format:  The file format to use for reading the array. Default is None - will be read by the file
                             extension.
        :param allow_pickle: Whether to allow loading pickled arrays in case of object type arrays. Only relevant to
                             'npy' format. Default is False for security reasons.

        :return: The unpacked array.
        """
        # Get the file:
        file_path = self.get_data_item_local_path(data_item=data_item)

        # Get the archive format by the file extension if needed:
        if file_format is None:
            file_format = NumPySupportedFormat.match_format(path=file_path)
        if (
            file_format is None
            or file_format in NumPySupportedFormat.get_multi_array_formats()
        ):
            raise MLRunInvalidArgumentError(
                f"File format of {data_item.key} ('{''.join(pathlib.Path(file_path).suffixes)}') is not supported. "
                f"Supported formats are: {' '.join(NumPySupportedFormat.get_single_array_formats())}"
            )

        # Read the object:
        formatter = NumPySupportedFormat.get_format_handler(fmt=file_format)
        load_kwargs = {}
        if file_format == NumPySupportedFormat.NPY:
            load_kwargs["allow_pickle"] = allow_pickle
        obj = formatter.load(file_path=file_path, **load_kwargs)

        return obj

    def unpack_dataset(self, data_item: DataItem) -> np.ndarray:
        """
        Unpack a numppy array from a dataset artifact.

        :param data_item: The data item to unpack.

        :return: The unpacked array.
        """
        # Get the artifact's data frame:
        data_frame = data_item.as_df()

        # Cast the data frame to a `np.ndarray` (1D arrays are returned as a 2D array with shape of 1xn, so we use
        # squeeze to decrease the redundant dimension):
        array = data_frame.to_numpy().squeeze()

        return array


class _NumPyNDArrayCollectionPackager(DefaultPackager):
    """
    A base packager for builtin python dictionaries and lists of numpy arrays as they share common artifact and file
    types.
    """

    DEFAULT_PACKING_ARTIFACT_TYPE = ArtifactType.FILE
    DEFAULT_UNPACKING_ARTIFACT_TYPE = ArtifactType.FILE
    PRIORITY = 4

    def pack_file(
        self,
        obj: NumPyArrayCollectionType,
        key: str,
        file_format: str = DEFAULT_NUMPPY_ARRAY_COLLECTION_FORMAT,
        **save_kwargs,
    ) -> tuple[Artifact, dict]:
        """
        Pack an array collection as a file by the given format.

        :param obj:         The aray collection to pack.
        :param key:         The key to use for the artifact.
        :param file_format: The file format to save as. Default is npy.
        :param save_kwargs: Additional keyword arguments to pass to the numpy save functions.

        :return: The packed artifact and instructions.
        """
        # Save to file:
        formatter = NumPySupportedFormat.get_format_handler(fmt=file_format)
        temp_directory = pathlib.Path(tempfile.mkdtemp())
        self.add_future_clearing_path(path=temp_directory)
        file_path = temp_directory / f"{key}.{file_format}"
        formatter.save(obj=obj, file_path=str(file_path), **save_kwargs)

        # Create the artifact and instructions (Note: only 'npz' format support saving object arrays and that will
        # require pickling, hence we set the required instruction):
        artifact = Artifact(key=key, src_path=os.path.abspath(file_path))
        instructions = {"file_format": file_format}
        if file_format == NumPySupportedFormat.NPZ and self._is_any_object_dtype(
            array_collection=obj
        ):
            instructions["allow_pickle"] = True

        return artifact, instructions

    def unpack_file(
        self,
        data_item: DataItem,
        file_format: Optional[str] = None,
        allow_pickle: bool = False,
    ) -> dict[str, np.ndarray]:
        """
        Unpack a numppy array collection from file.

        :param data_item:    The data item to unpack.
        :param file_format:  The file format to use for reading the array collection. Default is None - will be read by
                             the file extension.
        :param allow_pickle: Whether to allow loading pickled arrays in case of object type arrays. Only relevant to
                             'npz' format. Default is False for security reasons.

        :return: The unpacked array collection.
        """
        # Get the file:
        file_path = self.get_data_item_local_path(data_item=data_item)

        # Get the archive format by the file extension if needed:
        if file_format is None:
            file_format = NumPySupportedFormat.match_format(path=file_path)
        if (
            file_format is None
            or file_format in NumPySupportedFormat.get_single_array_formats()
        ):
            raise MLRunInvalidArgumentError(
                f"File format of {data_item.key} ('{''.join(pathlib.Path(file_path).suffixes)}') is not supported. "
                f"Supported formats are: {' '.join(NumPySupportedFormat.get_multi_array_formats())}"
            )

        # Read the object:
        formatter = NumPySupportedFormat.get_format_handler(fmt=file_format)
        load_kwargs = {}
        if file_format == NumPySupportedFormat.NPZ:
            load_kwargs["allow_pickle"] = allow_pickle
        obj = formatter.load(file_path=file_path, **load_kwargs)

        return obj

    @staticmethod
    def _is_any_object_dtype(
        array_collection: Union[np.ndarray, NumPyArrayCollectionType],
    ):
        """
        Check if any of the arrays in a collection is of type `object`.

        :param array_collection: The collection to check fo `object` dtype.

        :return: True if at least one array in the collection is an `object` array.
        """
        if isinstance(array_collection, list):
            return any(
                _NumPyNDArrayCollectionPackager._is_any_object_dtype(
                    array_collection=array
                )
                for array in array_collection
            )
        elif isinstance(array_collection, dict):
            return any(
                _NumPyNDArrayCollectionPackager._is_any_object_dtype(
                    array_collection=array
                )
                for array in array_collection.values()
            )
        return array_collection.dtype == np.object_


class NumPyNDArrayDictPackager(_NumPyNDArrayCollectionPackager):
    """
    ``dict[str, numpy.ndarray]`` packager.
    """

    PACKABLE_OBJECT_TYPE = dict[str, np.ndarray]

    def is_packable(
        self,
        obj: Any,
        artifact_type: Optional[str] = None,
        configurations: Optional[dict] = None,
    ) -> bool:
        """
        Check if the object provided is a dictionary of numpy arrays.

        :param obj:            The object to pack.
        :param artifact_type:  The artifact type to log the object as.
        :param configurations: The log hint configurations passed by the user.

        :return: True if packable and False otherwise.
        """
        # Check the obj is a dictionary with string keys and arrays as values:
        if not (
            isinstance(obj, dict)
            and all(
                isinstance(key, str) and isinstance(value, np.ndarray)
                for key, value in obj.items()
            )
        ):
            return False

        # Check the artifact type is supported:
        if artifact_type and artifact_type not in self.get_supported_artifact_types():
            return False

        # Check an edge case where the dictionary is empty (this packager will pack empty dictionaries only if given
        # specific file format, otherwise it will be packed by the `DictPackager`):
        if not obj:
            return (
                configurations.get("file_format", None)
                in NumPySupportedFormat().get_multi_array_formats()
            )

        return True

    def pack_result(self, obj: dict[str, np.ndarray], key: str) -> dict:
        """
        Pack a dictionary of numpy arrays as a result.

        :param obj: The arrays dictionary to pack and log.
        :param key: The result's key.

        :return: The result dictionary.
        """
        return {
            key: {
                array_key: array_value.tolist()
                for array_key, array_value in obj.items()
            }
        }

    def unpack_file(
        self,
        data_item: DataItem,
        file_format: Optional[str] = None,
        allow_pickle: bool = False,
    ) -> dict[str, np.ndarray]:
        """
        Unpack a numppy array dictionary from file.

        :param data_item:    The data item to unpack.
        :param file_format:  The file format to use for reading the arrays dictionary. Default is None - will be read by
                             the file extension.
        :param allow_pickle: Whether to allow loading pickled arrays in case of object type arrays. Only relevant to
                             'npz' format. Default is False for security reasons.

        :return: The unpacked array.
        """
        # Load the object:
        obj = super().unpack_file(
            data_item=data_item, file_format=file_format, allow_pickle=allow_pickle
        )

        # The returned object is a mapping of type NpzFile, so we cast it to a dictionary:
        return {key: array for key, array in obj.items()}


class NumPyNDArrayListPackager(_NumPyNDArrayCollectionPackager):
    """
    ``list[numpy.ndarray]`` packager.
    """

    PACKABLE_OBJECT_TYPE = list[np.ndarray]

    def is_packable(
        self,
        obj: Any,
        artifact_type: Optional[str] = None,
        configurations: Optional[dict] = None,
    ) -> bool:
        """
        Check if the object provided is a list of numpy arrays.

        :param obj:            The object to pack.
        :param artifact_type:  The artifact type to log the object as.
        :param configurations: The log hint configurations passed by the user.

        :return: True if packable and False otherwise.
        """
        # Check the obj is a list with arrays as values:
        if not (
            isinstance(obj, list)
            and all(isinstance(value, np.ndarray) for value in obj)
        ):
            return False

        # Check the artifact type is supported:
        if artifact_type and artifact_type not in self.get_supported_artifact_types():
            return False

        # Check an edge case where the list is empty (this packager will pack empty lists only if given specific file
        # format, otherwise it will be packed by the `ListPackager`):
        if not obj:
            return (
                configurations.get("file_format", None)
                in NumPySupportedFormat().get_multi_array_formats()
            )

        return True

    def pack_result(self, obj: list[np.ndarray], key: str) -> dict:
        """
        Pack a list of numpy arrays as a result.

        :param obj: The arrays list to pack and log.
        :param key: The result's key.

        :return: The result dictionary.
        """
        return {key: [array.tolist() for array in obj]}

    def unpack_file(
        self,
        data_item: DataItem,
        file_format: Optional[str] = None,
        allow_pickle: bool = False,
    ) -> list[np.ndarray]:
        """
        Unpack a numppy array list from file.

        :param data_item:    The data item to unpack.
        :param file_format:  The file format to use for reading the arrays list. Default is None - will be read by the
                             file extension.
        :param allow_pickle: Whether to allow loading pickled arrays in case of object type arrays. Only relevant to
                             'npz' format. Default is False for security reasons.

        :return: The unpacked array.
        """
        # Load the object:
        obj = super().unpack_file(
            data_item=data_item, file_format=file_format, allow_pickle=allow_pickle
        )

        # The returned object is a mapping of type NpzFile, so we cast it to a list:
        return list(obj.values())


class NumPyNumberPackager(DefaultPackager):
    """
    ``numpy.number`` packager. It is also used for all `number` inheriting numpy objects (`float32`, uint8, etc.).
    """

    PACKABLE_OBJECT_TYPE = np.number
    DEFAULT_PACKING_ARTIFACT_TYPE = ArtifactType.RESULT
    PACK_SUBCLASSES = True  # To include all dtypes ('float32', 'uint8', ...)

    def pack_result(self, obj: np.number, key: str) -> dict:
        """
        Pack a numpy number as a result.

        :param obj: The number to pack and log.
        :param key: The result's key.

        :return: The result dictionary.
        """
        return super().pack_result(obj=obj.item(), key=key)



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC
from typing import Generic, TypeVar, Union

# A generic type for a supported format handler class type:
FileHandlerType = TypeVar("FileHandlerType")


class SupportedFormat(ABC, Generic[FileHandlerType]):
    """
    Library of supported formats by some builtin MLRun packagers.
    """

    # Add here all the supported formats in ALL CAPS and their value as a string:
    ...

    # The map to use in the method `get_format_handler`. A dictionary of string key to a class type to handle that
    # format. New supported formats and handlers should be added to it:
    _FORMAT_HANDLERS_MAP: dict[str, type[FileHandlerType]] = {}

    @classmethod
    def get_all_formats(cls) -> list[str]:
        """
        Get all supported formats.

        :return: A list of all the supported formats.
        """
        return [
            value
            for key, value in cls.__dict__.items()
            if isinstance(value, str) and not key.startswith("_")
        ]

    @classmethod
    def get_format_handler(cls, fmt: str) -> type[FileHandlerType]:
        """
        Get the format handler to the provided format (file extension):

        :param fmt: The file extension to get the corresponding handler.

        :return: The handler class.
        """
        return cls._FORMAT_HANDLERS_MAP[fmt]

    @classmethod
    def match_format(cls, path: str) -> Union[str, None]:
        """
        Try to match one of the available formats this class holds to a given path.

        :param path: The path to match the format to.

        :return: The matched format if found and None otherwise.
        """
        formats = cls.get_all_formats()
        for fmt in formats:
            if path.endswith(f".{fmt}"):
                return fmt
        return None



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = [
    "get_version",
    "set_environment",
    "code_to_function",
    "import_function",
    "handler",
    "ArtifactType",
    "get_secret_or_env",
    "mount_v3io",
    "v3io_cred",
    "auto_mount",
    "VolumeMount",
]

from os import environ, path
from typing import Optional

import dotenv

from .common.constants import MLRUN_ACTIVE_PROJECT
from .config import config as mlconf
from .datastore import DataItem, ModelProvider, store_manager
from .db import get_run_db
from .errors import MLRunInvalidArgumentError, MLRunNotFoundError
from .execution import MLClientCtx
from .hub import get_hub_module, import_module
from .model import RunObject, RunTemplate, new_task
from .package import ArtifactType, DefaultPackager, Packager, handler
from .projects import (
    MlrunProject,
    ProjectMetadata,
    build_function,
    deploy_function,
    get_or_create_project,
    load_project,
    new_project,
    pipeline_context,
    run_function,
)
from .projects.project import _add_username_to_project_name_if_needed
from .run import (
    _run_pipeline,
    code_to_function,
    function_to_module,
    get_dataitem,
    get_model_provider,
    get_object,
    get_or_create_ctx,
    get_pipeline,
    import_function,
    new_function,
    retry_pipeline,
    terminate_pipeline,
    wait_for_pipeline_completion,
)
from .runtimes import mounts, new_model_server
from .secrets import get_secret_or_env
from .utils.version import Version

__version__ = Version().get()["version"]

VolumeMount = mounts.VolumeMount
mount_v3io = mounts.mount_v3io
v3io_cred = mounts.v3io_cred
auto_mount = mounts.auto_mount


def get_version():
    """get current mlrun version"""
    return __version__


if "IGZ_NAMESPACE_DOMAIN" in environ:
    igz_domain = environ["IGZ_NAMESPACE_DOMAIN"]
    kfp_ep = f"https://dashboard.{igz_domain}/pipelines"
    environ["KF_PIPELINES_UI_ENDPOINT"] = kfp_ep
    mlconf.remote_host = mlconf.remote_host or igz_domain


def set_environment(
    api_path: Optional[str] = None,
    artifact_path: str = "",
    access_key: Optional[str] = None,
    username: Optional[str] = None,
    env_file: Optional[str] = None,
    mock_functions: Optional[str] = None,
):
    """set and test default config for: api path, artifact_path and project

    this function will try and read the configuration from the environment/api
    and merge it with the user provided project name, artifacts path or api path/access_key.
    it returns the configured artifacts path, this can be used to define sub paths.

    Note: the artifact path is an mlrun data uri (e.g. `s3://bucket/path`) and can not be used with file utils.

    example::

        from os import path

        project_name, artifact_path = set_environment()
        set_environment("http://localhost:8080", artifact_path="./")
        set_environment(env_file="mlrun.env")
        set_environment("<remote-service-url>", access_key="xyz", username="joe")

    :param api_path:       location/url of mlrun api service
    :param artifact_path:  path/url for storing experiment artifacts
    :param access_key:     set the remote cluster access key (V3IO_ACCESS_KEY)
    :param username:       name of the user to authenticate
    :param env_file:       path/url to .env file (holding MLRun config and other env vars), see: set_env_from_file()
    :param mock_functions: set to True to create local/mock functions instead of real containers,
                           set to "auto" to auto determine based on the presence of k8s/Nuclio
    :returns:
        active project name
        actual artifact path/url, can be used to create subpaths per task or group of artifacts
    """
    if env_file:
        set_env_from_file(env_file)

    # set before the dbpath (so it will re-connect with the new credentials)
    if access_key:
        environ["V3IO_ACCESS_KEY"] = access_key
    if username:
        environ["V3IO_USERNAME"] = username

    mlconf.dbpath = api_path or mlconf.dbpath
    if not mlconf.dbpath:
        raise ValueError("DB/API path was not detected, please specify its address")

    # check connectivity and load remote defaults
    get_run_db()
    if api_path:
        environ["MLRUN_DBPATH"] = mlconf.dbpath
    mlconf.reload()

    if mock_functions is not None:
        mock_functions = "1" if mock_functions is True else mock_functions
        mlconf.force_run_local = mock_functions
        mlconf.mock_nuclio_deployment = mock_functions

    if not mlconf.artifact_path and not artifact_path:
        raise ValueError(
            "default artifact_path was not configured, please specify a valid artifact_path"
        )

    if artifact_path:
        if artifact_path.startswith("./"):
            artifact_path = path.abspath(artifact_path)
        elif not artifact_path.startswith("/") and "://" not in artifact_path:
            raise ValueError(
                "artifact_path must refer to an absolute path" " or a valid url"
            )
        mlconf.artifact_path = artifact_path

    return mlconf.active_project, mlconf.artifact_path


def get_current_project(silent: bool = False) -> Optional[MlrunProject]:
    if pipeline_context.project:
        return pipeline_context.project

    project_name = environ.get(MLRUN_ACTIVE_PROJECT, None)
    if not project_name:
        if not silent:
            raise MLRunInvalidArgumentError(
                "No current project is initialized. Use new, get or load project functions first."
            )
        return None

    project = load_project(
        name=project_name,
        url=project_name,
        save=False,
        sync_functions=False,
    )

    if not project and not silent:
        raise MLRunInvalidArgumentError(
            "No current project is initialized. Use new, get or load project functions first."
        )
    return project


def get_sample_path(subpath=""):
    """
    return the url of a sample dataset or model
    """
    samples_path = environ.get(
        "SAMPLE_DATA_SOURCE_URL_PREFIX", mlconf.default_samples_path
    )
    if subpath:
        samples_path = path.join(samples_path, subpath.lstrip("/"))
    return samples_path


def set_env_from_file(env_file: str, return_dict: bool = False) -> Optional[dict]:
    """Read and set and/or return environment variables from a file
    the env file should have lines in the form KEY=VALUE, comment line start with "#"

    example file::

        # this is an env file
        MLRUN_DBPATH=https://mlrun-api.default-tenant.app.xxx.iguazio-cd1.com
        V3IO_USERNAME=admin
        V3IO_API=https://webapi.default-tenant.app.xxx.iguazio-cd1.com
        V3IO_ACCESS_KEY=MYKEY123
        AWS_ACCESS_KEY_ID-XXXX
        AWS_SECRET_ACCESS_KEY=YYYY

    usage::

        # set the env vars from a file + return the results as a dict
        env_dict = mlrun.set_env_from_file(env_path, return_dict=True)

    :param env_file:    path/url to env file
    :param return_dict: set to True to return the env as a dict
    :return: None or env dict
    """
    env_file = path.expanduser(env_file)
    if not path.isfile(env_file):
        raise MLRunNotFoundError(f"env file {env_file} does not exist")
    env_vars = dotenv.dotenv_values(env_file)
    if None in env_vars.values():
        raise MLRunInvalidArgumentError("env file lines must be in the form key=value")

    for key, value in env_vars.items():
        environ[key] = value

    # reload mlrun configuration
    mlconf.reload()
    return env_vars if return_dict else None



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os
import urllib
from typing import Optional
from urllib.parse import urlparse

import requests
import v3io

import mlrun.errors
from mlrun.config import config as mlconf
from mlrun.utils import dict_to_json

_cached_control_session = None


def split_path(mntpath=""):
    if mntpath[0] == "/":
        mntpath = mntpath[1:]
    paths = mntpath.split("/")
    container = paths[0]
    subpath = ""
    if len(paths) > 1:
        subpath = mntpath[len(container) :]
    return container, subpath


def v3io_to_vol(name, remote="~/", access_key="", user="", secret=None):
    from os import environ

    from kubernetes import client

    access_key = access_key or environ.get("V3IO_ACCESS_KEY")
    opts = {"accessKey": access_key}
    user = user or environ.get("V3IO_USERNAME")
    if user:
        opts["dirsToCreate"] = f'[{{"name": "users//{user}", "permissions": 488}}]'

    remote = str(remote)

    if remote.startswith("~/"):
        if not user:
            raise mlrun.errors.MLRunInvalidArgumentError(
                'user name/env must be specified when using "~" in path'
            )
        if remote == "~/":
            remote = "users/" + user
        else:
            remote = "users/" + user + remote[1:]
    if remote:
        container, subpath = split_path(remote)
        opts["container"] = container
        opts["subPath"] = subpath

    if secret:
        secret = {"name": secret}

    vol = {
        "flexVolume": client.V1FlexVolumeSource(
            "v3io/fuse", options=opts, secret_ref=secret
        ),
        "name": name,
    }
    return vol


class OutputStream:
    def __init__(
        self,
        stream_path,
        shards=None,
        retention_in_hours=None,
        create=True,
        endpoint=None,
        access_key=None,
        mock=False,
        **kwargs,  # to avoid failing on extra parameters
    ):
        v3io_client_kwargs = {}
        if endpoint:
            v3io_client_kwargs["endpoint"] = endpoint
        if access_key:
            v3io_client_kwargs["access_key"] = access_key

        if not mock:
            self._v3io_client = v3io.dataplane.Client(**v3io_client_kwargs)
        else:
            self._v3io_client = None

        self._container, self._stream_path = split_path(stream_path)
        self._shards = shards
        self._retention_in_hours = retention_in_hours
        self._create = create
        self._endpoint = endpoint
        self._mock = mock
        self._mock_queue = []

    def create_stream(self) -> None:
        # this import creates an import loop via the utils module, so putting it in execution path
        from mlrun.utils.helpers import logger

        logger.debug(
            "Creating output stream",
            endpoint=self._endpoint,
            container=self._container,
            stream_path=self._stream_path,
            shards=self._shards,
            retention_in_hours=self._retention_in_hours,
        )
        response = self._v3io_client.stream.create(
            container=self._container,
            stream_path=self._stream_path,
            shard_count=self._shards or 1,
            retention_period_hours=self._retention_in_hours or 24,
            raise_for_status=v3io.dataplane.RaiseForStatus.never,
        )
        if not (response.status_code == 400 and "ResourceInUse" in str(response.body)):
            response.raise_for_status([409, 204])

    def _lazy_init(self):
        if self._create and not self._mock:
            self._create = False
            self.create_stream()

    def push(self, data, partition_key=None):
        self._lazy_init()

        def dump_record(rec):
            if not isinstance(rec, (str, bytes)):
                return dict_to_json(rec)
            return str(rec)

        if not isinstance(data, list):
            data = [data]

        records = []
        for rec in data:
            record = {"data": dump_record(rec)}
            if partition_key is not None:
                record["partition_key"] = partition_key
            records.append(record)

        if self._mock:
            # for mock testing
            self._mock_queue.extend(records)
        else:
            self._v3io_client.stream.put_records(
                container=self._container,
                stream_path=self._stream_path,
                records=records,
            )


class HTTPOutputStream:
    """HTTP output source that usually used for CE mode and debugging process"""

    def __init__(self, stream_path: str):
        self._stream_path = stream_path

    def push(self, data):
        def dump_record(rec):
            if isinstance(rec, bytes):
                return rec

            if not isinstance(rec, str):
                rec = dict_to_json(rec)

            return rec.encode("UTF-8")

        if not isinstance(data, list):
            data = [data]

        for record in data:
            # Convert the new record to the required format
            serialized_record = dump_record(record)
            response = requests.post(self._stream_path, data=serialized_record)
            if not response:
                raise mlrun.errors.MLRunInvalidArgumentError(
                    f"API call failed push a new record through {self._stream_path}, "
                    f"status {response.status_code}: {response.reason}"
                )


class KafkaOutputStream:
    def __init__(
        self,
        topic,
        brokers,
        producer_options=None,
        mock=False,
    ):
        self._kafka_producer = None
        self._topic = topic
        self._brokers = brokers
        self._producer_options = producer_options or {}

        self._mock = mock
        self._mock_queue = []

        self._initialized = False

    def _lazy_init(self):
        if self._initialized or self._mock:
            return

        import kafka

        self._kafka_producer = kafka.KafkaProducer(
            bootstrap_servers=self._brokers,
            **self._producer_options,
        )

        self._initialized = True

    def push(self, data, partition_key=None):
        self._lazy_init()

        def dump_record(rec):
            if isinstance(rec, bytes):
                return rec

            if not isinstance(rec, str):
                rec = dict_to_json(rec)

            return rec.encode("UTF-8")

        if not isinstance(data, list):
            data = [data]

        if self._mock:
            # for mock testing
            self._mock_queue.extend(data)
        else:
            for record in data:
                serialized_record = dump_record(record)
                if isinstance(partition_key, str):
                    partition_key = partition_key.encode("UTF-8")
                self._kafka_producer.send(
                    self._topic, serialized_record, key=partition_key
                )


class V3ioStreamClient:
    def __init__(
        self, url: str, shard_id: int = 0, seek_to: Optional[str] = None, **kwargs
    ):
        endpoint, stream_path = parse_path(url)
        seek_options = ["EARLIEST", "LATEST", "TIME", "SEQUENCE"]
        seek_to = seek_to or "LATEST"
        seek_to = seek_to.upper()
        if seek_to not in seek_options:
            raise ValueError(f'seek_to must be one of {", ".join(seek_options)}')

        self._url = url
        self._container, self._stream_path = split_path(stream_path)
        self._shard_id = shard_id
        self._seek_to = seek_to
        self._client = v3io.dataplane.Client(endpoint=endpoint)
        self._seek_done = False
        self._location = ""
        self._kwargs = kwargs

    @property
    def url(self):
        return self._url

    @property
    def shard_id(self):
        return self._shard_id

    def seek(self):
        response = self._client.stream.seek(
            self._container,
            self._stream_path,
            self._shard_id,
            self._seek_to,
            raise_for_status=v3io.dataplane.RaiseForStatus.never,
            **self._kwargs,
        )
        if response.status_code == 404 and "ResourceNotFound" in str(response.body):
            return 0
        response.raise_for_status()
        self._location = response.output.location
        self._seek_done = True
        return response.status_code

    def get_records(self):
        if not self._seek_done:
            resp = self.seek()
            if resp == 0:
                return []
        response = self._client.stream.get_records(
            self._container, self._stream_path, self._shard_id, self._location
        )
        response.raise_for_status()
        self._location = response.output.next_location
        return response.output.records


def is_iguazio_endpoint(endpoint_url: str) -> bool:
    # TODO: find a better heuristic
    return ".default-tenant." in endpoint_url


def is_iguazio_session(value: str) -> bool:
    # TODO: find a better heuristic
    return len(value) > 20 and "-" in value


def is_iguazio_session_cookie(session_cookie: str) -> bool:
    if not session_cookie.strip():
        return False

    # decode url encoded cookie
    # from: j%3A%7B%22sid%22%3A%20%22946b0749-5c40-4837-a4ac-341d295bfaf7%22%7D
    # to:   j:{"sid":"946b0749-5c40-4837-a4ac-341d295bfaf7"}
    try:
        unqouted_cookie = urllib.parse.unquote(session_cookie.strip())
        if not unqouted_cookie.startswith("j:"):
            return is_iguazio_session(session_cookie)
        return json.loads(unqouted_cookie[2:])["sid"] is not None
    except Exception:
        return False


# we assign the control session or access key to the password since this is iguazio auth scheme
# (requests should be sent with username:control_session/access_key as auth header)
def add_or_refresh_credentials(
    api_url: str, username: str = "", password: str = "", token: str = ""
) -> (str, str, str):
    if is_iguazio_session(password):
        return username, password, token

    username = username or os.environ.get("V3IO_USERNAME")
    password = password or os.environ.get("V3IO_PASSWORD")
    # V3IO_ACCESS_KEY` is used by other packages like v3io, MLRun also uses it as the access key used to
    # communicate with the API from the client. `MLRUN_AUTH_SESSION` is for when we want
    # different access keys for the 2 usages
    token = (
        token
        # can't use mlrun.common.runtimes.constants.FunctionEnvironmentVariables.auth_session cause this is running
        # in the import execution path (when we're initializing the run db) and therefore we can't import mlrun.runtimes
        or os.environ.get("MLRUN_AUTH_SESSION")
        or os.environ.get("V3IO_ACCESS_KEY")
    )

    # When it's not iguazio endpoint it's one of two options:
    # Enterprise, but we're in the cluster (and not from remote), e.g. url will be something like http://mlrun-api:8080
    # In which we enforce to have access key which is needed for the API auth
    # Open source in which auth is not enabled so no creds needed
    # We don't really have an easy/nice way to differentiate between the two so we're just sending creds anyways
    # (ideally if we could identify we're in enterprise we would have verify here that token and username have value)
    if not is_iguazio_endpoint(api_url):
        return "", "", token

    if not username or not token:
        raise ValueError(
            "username and access key required to authenticate against iguazio system"
        )
    return username, token, ""


def parse_path(url, suffix="/"):
    """return endpoint and table path from url"""
    parsed_url = urlparse(url)
    if parsed_url.netloc:
        scheme = parsed_url.scheme.lower()
        if scheme == "s3":
            prefix = "s3"
        elif scheme == "v3ios":
            prefix = "https"
        elif scheme == "v3io":
            prefix = "http"
        elif scheme == "redis":
            prefix = "redis"
        elif scheme == "rediss":
            prefix = "rediss"
        elif scheme == "ds":
            prefix = "ds"
        else:
            raise mlrun.errors.MLRunInvalidArgumentError(
                "url must start with v3io/v3ios/redis/rediss, got " + url
            )
        endpoint = f"{prefix}://{parsed_url.netloc}"
    else:
        # no netloc is mainly when using v3io (v3io:///) and expecting the url to be resolved automatically from env or
        # config
        endpoint = None
    return endpoint, parsed_url.path.strip("/") + suffix


def sanitize_username(username: str):
    """
    The only character an Iguazio username may have that is not valid for k8s usage is underscore (_)
    So simply replace it with dash
    """
    return username.replace("_", "-")


def min_iguazio_versions(*versions):
    def decorator(function):
        def wrapper(*args, **kwargs):
            if mlrun.utils.helpers.validate_component_version_compatibility(
                "iguazio", *versions
            ):
                return function(*args, **kwargs)

            message = (
                f"{function.__name__} is supported since Iguazio {' or '.join(versions)}, currently using "
                f"Iguazio {mlconf.igz_version}."
            )
            raise mlrun.errors.MLRunIncompatibleVersionError(message)

        return wrapper

    return decorator



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re

import pandas as pd

from .base import BaseMerger


class LocalFeatureMerger(BaseMerger):
    engine = "local"
    support_offline = True

    def __init__(self, vector, **engine_args):
        super().__init__(vector, **engine_args)

    def _asof_join(
        self,
        entity_df,
        entity_timestamp_column: str,
        featureset_name,
        featureset_timstamp,
        featureset_df,
        left_keys: list,
        right_keys: list,
    ):
        index_col_not_in_entity = "index" not in entity_df.columns
        index_col_not_in_featureset = "index" not in featureset_df.columns
        entity_df[entity_timestamp_column] = pd.to_datetime(
            entity_df[entity_timestamp_column]
        )
        featureset_df[featureset_timstamp] = pd.to_datetime(
            featureset_df[featureset_timstamp]
        )
        entity_df.sort_values(by=entity_timestamp_column, inplace=True)
        featureset_df.sort_values(by=featureset_timstamp, inplace=True)

        featureset_df = self._normalize_timestamp_column(
            entity_timestamp_column,
            entity_df,
            featureset_timstamp,
            featureset_df,
            featureset_name,
        )

        merged_df = pd.merge_asof(
            entity_df,
            featureset_df,
            left_on=entity_timestamp_column,
            right_on=featureset_timstamp,
            left_by=left_keys or None,
            right_by=right_keys or None,
            suffixes=("", f"_{featureset_name}_"),
        )
        for col in merged_df.columns:
            if re.findall(f"_{featureset_name}_$", col):
                self._append_drop_column(col)
        # Undo indexing tricks for asof merge
        # to return the correct indexes and not
        # overload `index` columns
        if (
            "index" not in left_keys
            and "index" not in right_keys
            and index_col_not_in_entity
            and index_col_not_in_featureset
            and "index" in merged_df.columns
        ):
            merged_df.drop(columns="index", inplace=True)
        return merged_df

    def _join(
        self,
        entity_df,
        entity_timestamp_column: str,
        featureset_name,
        featureset_timestamp,
        featureset_df,
        left_keys: list,
        right_keys: list,
    ):
        merged_df = pd.merge(
            entity_df,
            featureset_df,
            how=self._join_type,
            left_on=left_keys,
            right_on=right_keys,
            suffixes=("", f"_{featureset_name}_"),
        )
        for col in merged_df.columns:
            if re.findall(f"_{featureset_name}_$", col):
                self._append_drop_column(col)
        return merged_df

    def _create_engine_env(self):
        pass

    def _get_engine_df(
        self,
        feature_set,
        feature_set_name,
        column_names=None,
        start_time=None,
        end_time=None,
        time_column=None,
        additional_filters=None,
    ):
        df = feature_set.to_dataframe(
            columns=column_names,
            start_time=start_time,
            end_time=end_time,
            time_column=time_column,
            additional_filters=additional_filters,
        )
        if df.index.names[0]:
            df.reset_index(inplace=True)
        return df

    def _rename_columns_and_select(self, df, rename_col_dict, columns=None):
        df.rename(
            columns=rename_col_dict,
            inplace=True,
        )

    def _drop_columns_from_result(self):
        self._result_df.drop(columns=self._drop_columns, inplace=True, errors="ignore")

    def _filter(self, query):
        self._result_df.query(query, inplace=True)

    def _order_by(self, order_by_active):
        self._result_df.sort_values(by=order_by_active, ignore_index=True, inplace=True)

    def _convert_entity_rows_to_engine_df(self, entity_rows):
        return entity_rows



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
import os
import warnings
from base64 import b64decode
from copy import deepcopy
from typing import Optional, Union

import nuclio
from nuclio import KafkaTrigger

import mlrun
import mlrun.common.schemas as schemas
import mlrun.common.secrets
import mlrun.datastore.datastore_profile as ds_profile
from mlrun.datastore import get_kafka_brokers_from_dict, parse_kafka_url
from mlrun.model import ObjectList
from mlrun.runtimes.function_reference import FunctionReference
from mlrun.secrets import SecretsStore
from mlrun.serving.server import (
    GraphServer,
    add_system_steps_to_graph,
    create_graph_server,
)
from mlrun.serving.states import (
    RootFlowStep,
    RouterStep,
    StepKinds,
    TaskStep,
    graph_root_setter,
    new_remote_endpoint,
    params_to_step,
)
from mlrun.utils import get_caller_globals, logger, set_paths

from .. import KubejobRuntime
from ..pod import KubeResourceSpec
from .function import NuclioSpec, RemoteRuntime, min_nuclio_versions

serving_subkind = "serving_v2"


def new_v2_model_server(
    name,
    model_class: str,
    models: Optional[dict] = None,
    filename="",
    protocol="",
    image="",
    endpoint="",
    workers=8,
    canary=None,
):
    f = ServingRuntime()
    if not image:
        name, spec, code = nuclio.build_file(
            filename, name=name, handler="handler", kind=serving_subkind
        )
        f.spec.base_spec = spec

    f.metadata.name = name
    f.spec.default_class = model_class
    params = None
    if protocol:
        params = {"protocol": protocol}
    if models:
        for name, model_path in models.items():
            f.add_model(name, model_path=model_path, parameters=params)

    f.with_http(workers, host=endpoint, canary=canary)
    if image:
        f.from_image(image)

    return f


class ServingSpec(NuclioSpec):
    _dict_fields = NuclioSpec._dict_fields + [
        "graph",
        "load_mode",
        "graph_initializer",
        "function_refs",
        "parameters",
        "models",
        "default_content_type",
        "error_stream",
        "default_class",
        "secret_sources",
        "track_models",
    ]

    def __init__(
        self,
        command=None,
        args=None,
        image=None,
        mode=None,
        entry_points=None,
        description=None,
        replicas=None,
        min_replicas=None,
        max_replicas=None,
        volumes=None,
        volume_mounts=None,
        env=None,
        resources=None,
        config=None,
        base_spec=None,
        no_cache=None,
        source=None,
        image_pull_policy=None,
        function_kind=None,
        service_account=None,
        readiness_timeout=None,
        readiness_timeout_before_failure=None,
        models=None,
        graph=None,
        parameters=None,
        default_class=None,
        load_mode=None,
        build=None,
        function_refs=None,
        graph_initializer=None,
        error_stream=None,
        track_models=None,
        secret_sources=None,
        default_content_type=None,
        node_name=None,
        node_selector=None,
        affinity=None,
        disable_auto_mount=False,
        priority_class_name=None,
        default_handler=None,
        pythonpath=None,
        workdir=None,
        image_pull_secret=None,
        tolerations=None,
        preemption_mode=None,
        security_context=None,
        service_type=None,
        add_templated_ingress_host_mode=None,
        state_thresholds=None,
        disable_default_http_trigger=None,
        model_endpoint_creation_task_name=None,
        serving_spec=None,
    ):
        super().__init__(
            command=command,
            args=args,
            image=image,
            mode=mode,
            entry_points=entry_points,
            description=description,
            replicas=replicas,
            min_replicas=min_replicas,
            max_replicas=max_replicas,
            volumes=volumes,
            volume_mounts=volume_mounts,
            env=env,
            resources=resources,
            config=config,
            base_spec=base_spec,
            no_cache=no_cache,
            source=source,
            image_pull_policy=image_pull_policy,
            function_kind=serving_subkind,
            service_account=service_account,
            readiness_timeout=readiness_timeout,
            readiness_timeout_before_failure=readiness_timeout_before_failure,
            build=build,
            node_name=node_name,
            node_selector=node_selector,
            affinity=affinity,
            disable_auto_mount=disable_auto_mount,
            priority_class_name=priority_class_name,
            default_handler=default_handler,
            pythonpath=pythonpath,
            workdir=workdir,
            image_pull_secret=image_pull_secret,
            tolerations=tolerations,
            preemption_mode=preemption_mode,
            security_context=security_context,
            service_type=service_type,
            add_templated_ingress_host_mode=add_templated_ingress_host_mode,
            disable_default_http_trigger=disable_default_http_trigger,
            serving_spec=serving_spec,
        )

        self.models = models or {}
        self._graph = None
        self.graph: Union[RouterStep, RootFlowStep] = graph
        self.parameters = parameters or {}
        self.default_class = default_class
        self.load_mode = load_mode
        self._function_refs: ObjectList = None
        self.function_refs = function_refs or []
        self.graph_initializer = graph_initializer
        self.error_stream = error_stream
        self.track_models = track_models
        self.secret_sources = secret_sources or []
        self.default_content_type = default_content_type
        self.model_endpoint_creation_task_name = model_endpoint_creation_task_name

    @property
    def graph(self) -> Union[RouterStep, RootFlowStep]:
        """states graph, holding the serving workflow/DAG topology"""
        return self._graph

    @graph.setter
    def graph(self, graph):
        graph_root_setter(self, graph)

    @property
    def function_refs(self) -> list[FunctionReference]:
        """function references, list of optional child function refs"""
        return self._function_refs

    @function_refs.setter
    def function_refs(self, function_refs: list[FunctionReference]):
        self._function_refs = ObjectList.from_list(FunctionReference, function_refs)


class ServingRuntime(RemoteRuntime):
    """MLRun Serving Runtime"""

    kind = "serving"

    @property
    def spec(self) -> ServingSpec:
        return self._spec

    @spec.setter
    def spec(self, spec):
        self._spec = self._verify_dict(spec, "spec", ServingSpec)

    def set_topology(
        self,
        topology=None,
        class_name=None,
        engine=None,
        exist_ok=False,
        **class_args,
    ) -> Union[RootFlowStep, RouterStep]:
        """set the serving graph topology (router/flow) and root class or params

        examples::

            # simple model router topology
            graph = fn.set_topology("router")
            fn.add_model(name, class_name="ClassifierModel", model_path=model_uri)

            # async flow topology
            graph = fn.set_topology("flow", engine="async")
            graph.to("MyClass").to(name="to_json", handler="json.dumps").respond()

        topology options are::

          router - root router + multiple child route states/models
                   route is usually determined by the path (route key/name)
                   can specify special router class and router arguments

          flow   - workflow (DAG) with a chain of states
                   flow supports both "sync" and "async" engines, with "async" being the default.
                   Branches are not allowed in sync mode.
                   when using async mode calling state.respond() will mark the state as the
                   one which generates the (REST) call response

        :param topology:     - graph topology, router or flow
        :param class_name:   - optional for router, router class name/path or router object
        :param engine:       - optional for flow, sync or async engine
        :param exist_ok:     - allow overriding existing topology
        :param class_args:   - optional, router/flow class init args

        :return graph object (fn.spec.graph)
        """
        topology = topology or StepKinds.router
        if self.spec.graph and not exist_ok:
            raise mlrun.errors.MLRunInvalidArgumentError(
                "graph topology is already set, cannot be overwritten"
            )

        if topology == StepKinds.router:
            if class_name and hasattr(class_name, "to_dict"):
                _, step = params_to_step(class_name, None)
                if step.kind != StepKinds.router:
                    raise mlrun.errors.MLRunInvalidArgumentError(
                        "provided class is not a router step, must provide a router class in router topology"
                    )
            else:
                step = RouterStep(class_name=class_name, class_args=class_args)
            self.spec.graph = step
        elif topology == StepKinds.flow:
            self.spec.graph = RootFlowStep(engine=engine or "async")
            self.spec.graph.track_models = self.spec.track_models
        else:
            raise mlrun.errors.MLRunInvalidArgumentError(
                f"unsupported topology {topology}, use 'router' or 'flow'"
            )
        return self.spec.graph

    def set_tracking(
        self,
        stream_path: Optional[str] = None,
        batch: Optional[int] = None,
        sampling_percentage: float = 100,
        stream_args: Optional[dict] = None,
        enable_tracking: bool = True,
    ) -> None:
        """Apply on your serving function to monitor a deployed model, including real-time dashboards to detect drift
        and analyze performance.

        :param stream_path:                Path/url of the tracking stream e.g. v3io:///users/mike/mystream
                                           you can use the "dummy://" path for test/simulation.
        :param batch:                      Deprecated. Micro batch size (send micro batches of N records at a time).
        :param sampling_percentage:        Down sampling events that will be pushed to the monitoring stream based on
                                           a specified percentage. e.g. 50 for 50%. By default, all events are pushed.
        :param stream_args:                Stream initialization parameters, e.g. shards, retention_in_hours, ..
        :param enable_tracking:            Enabled/Disable model-monitoring tracking. Default True (tracking enabled).

        Example::

            # initialize a new serving function
            serving_fn = mlrun.import_function("hub://v2-model-server", new_name="serving")
            # apply model monitoring
            serving_fn.set_tracking()

        """
        # Applying model monitoring configurations
        self.spec.track_models = enable_tracking
        if self.spec.graph and isinstance(self.spec.graph, RootFlowStep):
            self.spec.graph.track_models = enable_tracking
        if self._spec and self._spec.function_refs:
            logger.debug(
                "Set tracking for children references", enable_tracking=enable_tracking
            )
            for name in self._spec.function_refs.keys():
                self._spec.function_refs[name].track_models = enable_tracking
                # Check if function_refs _function is filled if so update track_models field:
                if self._spec.function_refs[name]._function:
                    self._spec.function_refs[
                        name
                    ]._function.spec.track_models = enable_tracking

                    if self._spec.function_refs[
                        name
                    ]._function.spec.graph and isinstance(
                        self._spec.function_refs[name]._function.spec.graph,
                        RootFlowStep,
                    ):
                        self._spec.function_refs[
                            name
                        ]._function.spec.graph.track_models = enable_tracking

        if not 0 < sampling_percentage <= 100:
            raise mlrun.errors.MLRunInvalidArgumentError(
                "`sampling_percentage` must be greater than 0 and less or equal to 100."
            )
        self.spec.parameters["sampling_percentage"] = sampling_percentage

        if stream_path:
            self.spec.parameters["log_stream"] = stream_path
        if batch:
            warnings.warn(
                "The `batch` size parameter was deprecated in version 1.8.0 and is no longer used. "
                "It will be removed in 1.11.",
                # TODO: Remove this in 1.11
                FutureWarning,
            )
        if stream_args:
            self.spec.parameters["stream_args"] = stream_args

    def add_model(
        self,
        key: str,
        model_path: Optional[str] = None,
        class_name: Optional[str] = None,
        model_url: Optional[str] = None,
        handler: Optional[str] = None,
        router_step: Optional[str] = None,
        child_function: Optional[str] = None,
        creation_strategy: Optional[
            schemas.ModelEndpointCreationStrategy
        ] = schemas.ModelEndpointCreationStrategy.INPLACE,
        outputs: Optional[list[str]] = None,
        **class_args,
    ):
        """add ml model and/or route to the function.

        Example, create a function (from the notebook), add a model class, and deploy::

            fn = code_to_function(kind="serving")
            fn.add_model("boost", model_path, model_class="MyClass", my_arg=5)
            fn.deploy()

        only works with router topology, for nested topologies (model under router under flow)
        need to add router to flow and use router.add_route()

        :param key:         model api key (or name:version), will determine the relative url/path
        :param model_path:  path to mlrun model artifact or model directory file/object path
        :param class_name:  V2 Model python class name or a model class instance
                            (can also module.submodule.class and it will be imported automatically)
        :param model_url:   url of a remote model serving endpoint (cannot be used with model_path)
        :param handler:     for advanced users!, override default class handler name (do_event)
        :param router_step: router step name (to determine which router we add the model to in graphs
                            with multiple router steps)
        :param child_function: child function name, when the model runs in a child function
        :param creation_strategy: Strategy for creating or updating the model endpoint:
            * **overwrite**:
            1. If model endpoints with the same name exist, delete the `latest` one.
            2. Create a new model endpoint entry and set it as `latest`.
            * **inplace** (default):
            1. If model endpoints with the same name exist, update the `latest` entry.
            2. Otherwise, create a new entry.
            * **archive**:
            1. If model endpoints with the same name exist, preserve them.
            2. Create a new model endpoint with the same name and set it to `latest`.
        :param outputs: list of the model outputs (e.g. labels) ,if provided will override the outputs that been
                        configured in the model artifact, please note that those outputs need to be equal to the
                        model serving function outputs (length, and order)
        :param class_args:  extra kwargs to pass to the model serving class __init__
                            (can be read in the model using .get_param(key) method)
        """
        graph = self.spec.graph
        if not graph:
            graph = self.set_topology()

        if graph.kind != StepKinds.router:
            if router_step:
                if router_step not in graph:
                    raise ValueError(
                        f"router step {router_step} not present in the graph"
                    )
                graph = graph[router_step]
            else:
                routers = [
                    step
                    for step in graph.steps.values()
                    if step.kind == StepKinds.router
                ]
                if len(routers) == 0:
                    raise ValueError(
                        "graph does not contain any router, add_model can only be "
                        "used when there is a router step"
                    )
                if len(routers) > 1:
                    raise ValueError(
                        f"found {len(routers)} routers, please specify the router_step"
                        " you would like to add this model to"
                    )
                graph = routers[0]

        if class_name and hasattr(class_name, "to_dict"):
            if model_path:
                class_name.model_path = model_path
            if outputs:
                class_name.outputs = outputs
            key, state = params_to_step(
                class_name,
                key,
                model_endpoint_creation_strategy=creation_strategy,
                endpoint_type=schemas.EndpointType.LEAF_EP,
            )
        else:
            class_name = class_name or self.spec.default_class
            if class_name and not isinstance(class_name, str):
                raise ValueError(
                    "class name must be a string (name of module.submodule.name)"
                )
            if model_path and not class_name:
                raise ValueError("model_path must be provided with class_name")
            if model_path:
                model_path = str(model_path)

            if model_url:
                state = new_remote_endpoint(
                    model_url,
                    creation_strategy=creation_strategy,
                    endpoint_type=schemas.EndpointType.LEAF_EP,
                    **class_args,
                )
            else:
                class_args = deepcopy(class_args)
                class_args["model_path"] = model_path
                class_args["outputs"] = outputs
                state = TaskStep(
                    class_name,
                    class_args,
                    name=key,
                    handler=handler,
                    function=child_function,
                    model_endpoint_creation_strategy=creation_strategy,
                    endpoint_type=schemas.EndpointType.LEAF_EP,
                )

        return graph.add_route(key, state)

    def add_child_function(
        self, name, url=None, image=None, requirements=None, kind=None
    ):
        """in a multi-function pipeline add child function

        example::

            fn.add_child_function("enrich", "./enrich.ipynb", "mlrun/mlrun")

        :param name:   child function name
        :param url:    function/code url, support .py, .ipynb, .yaml extensions
        :param image:  base docker image for the function
        :param requirements: py package requirements file path OR list of packages
        :param kind:   mlrun function/runtime kind

        :return function object
        """
        function_reference = FunctionReference(
            url,
            image,
            requirements=requirements,
            kind=kind or "serving",
            track_models=self.spec.track_models,
        )
        self._spec.function_refs.update(function_reference, name)
        func = function_reference.to_function(self.kind)
        return func

    def _add_ref_triggers(self):
        """add stream trigger to downstream child functions"""
        for function_name, stream in self.spec.graph.get_queue_links().items():
            if stream.path:
                if function_name not in self._spec.function_refs.keys():
                    raise ValueError(f"function reference {function_name} not present")
                group = stream.options.get("group", f"{function_name}-consumer-group")

                child_function = self._spec.function_refs[function_name]
                trigger_args = stream.trigger_args or {}

                engine = self.spec.graph.engine or "async"
                if mlrun.mlconf.is_explicit_ack_enabled() and engine == "async":
                    trigger_args["explicit_ack_mode"] = trigger_args.get(
                        "explicit_ack_mode", "explicitOnly"
                    )
                    extra_attributes = trigger_args.get("extra_attributes", {})
                    trigger_args["extra_attributes"] = extra_attributes
                    extra_attributes["worker_allocation_mode"] = extra_attributes.get(
                        "worker_allocation_mode", "static"
                    )

                brokers = get_kafka_brokers_from_dict(stream.options)
                if stream.path.startswith("kafka://") or brokers:
                    if brokers:
                        brokers = brokers.split(",")
                    topic, brokers = parse_kafka_url(stream.path, brokers)
                    trigger = KafkaTrigger(
                        brokers=brokers,
                        topics=[topic],
                        consumer_group=group,
                        **trigger_args,
                    )
                    child_function.function_object.add_trigger("kafka", trigger)
                else:
                    # V3IO doesn't allow hyphens in object names
                    group = group.replace("-", "_")
                    child_function.function_object.add_v3io_stream_trigger(
                        stream.path, group=group, shards=stream.shards, **trigger_args
                    )

    def _deploy_function_refs(self, builder_env: Optional[dict] = None):
        """set metadata and deploy child functions"""
        for function_ref in self._spec.function_refs.values():
            logger.info(f"deploy child function {function_ref.name} ...")
            function_object = function_ref.function_object
            if not function_object:
                function_object = function_ref.to_function(self.kind)
            function_object.metadata.name = function_ref.fullname(self)
            function_object.metadata.project = self.metadata.project
            function_object.metadata.tag = self.metadata.tag

            function_object.metadata.labels = function_object.metadata.labels or {}
            function_object.metadata.labels["mlrun/parent-function"] = (
                self.metadata.name
            )
            function_object._is_child_function = True
            if not function_object.spec.graph:
                # copy the current graph only if the child doesnt have a graph of his own
                function_object.set_env("SERVING_CURRENT_FUNCTION", function_ref.name)
                function_object.spec.graph = self.spec.graph

            function_object.verbose = self.verbose
            function_object.spec.secret_sources = self.spec.secret_sources
            function_object.deploy(builder_env=builder_env)

    def remove_states(self, keys: list):
        """remove one, multiple, or all states/models from the spec (blank list for all)"""
        if self.spec.graph:
            self.spec.graph.clear_children(keys)

    def with_secrets(self, kind, source):
        """register a secrets source (file, env or dict)

        read secrets from a source provider to be used in workflows, example::

            task.with_secrets('file', 'file.txt')
            task.with_secrets('inline', {'key': 'val'})
            task.with_secrets('env', 'ENV1,ENV2')
            task.with_secrets('vault', ['secret1', 'secret2'...])

            # If using an empty secrets list [] then all accessible secrets will be available.
            task.with_secrets('vault', [])

            # To use with Azure key vault, a k8s secret must be created with the following keys:
            # kubectl -n <namespace> create secret generic azure-key-vault-secret \\
            #     --from-literal=tenant_id=<service principal tenant ID> \\
            #     --from-literal=client_id=<service principal client ID> \\
            #     --from-literal=secret=<service principal secret key>

            task.with_secrets('azure_vault', {
                'name': 'my-vault-name',
                'k8s_secret': 'azure-key-vault-secret',
                # An empty secrets list may be passed ('secrets': []) to access all vault secrets.
                'secrets': ['secret1', 'secret2'...]
            })

        :param kind:   secret type (file, inline, env)
        :param source: secret data or link (see example)

        :returns: The Runtime (function) object
        """
        if kind == "azure_vault" and isinstance(source, dict):
            candidate_secret_name = (source.get("k8s_secret") or "").strip()
            if candidate_secret_name:
                mlrun.common.secrets.validate_not_forbidden_secret(
                    candidate_secret_name
                )
        if kind == "vault" and isinstance(source, list):
            source = {"project": self.metadata.project, "secrets": source}

        self.spec.secret_sources.append({"kind": kind, "source": source})
        return self

    @min_nuclio_versions("1.12.10")
    def deploy(
        self,
        project="",
        tag="",
        verbose=False,
        builder_env: Optional[dict] = None,
        force_build: bool = False,
    ):
        """deploy model serving function to a local/remote cluster

        :param project:   optional, override function specified project name
        :param tag:       specify unique function tag (a different function service is created for every tag)
        :param verbose:   verbose logging
        :param builder_env: env vars dict for source archive config/credentials e.g. builder_env={"GIT_TOKEN": token}
        :param force_build: set True for force building the image
        """
        # Validate function name before deploying to k8s
        mlrun.utils.helpers.validate_function_name(self.metadata.name)

        load_mode = self.spec.load_mode
        if load_mode and load_mode not in ["sync", "async"]:
            raise ValueError(f"illegal model loading mode {load_mode}")
        if not self.spec.graph:
            raise ValueError("nothing to deploy, .spec.graph is none, use .add_model()")

        if self.spec.graph.kind != StepKinds.router and not getattr(
            self, "_is_child_function", None
        ):
            # initialize or create required streams/queues
            self.spec.graph.check_and_process_graph()
            self.spec.graph.create_queue_streams()
            functions_in_steps = self.spec.graph.list_child_functions()
            child_functions = list(self._spec.function_refs.keys())
            for function in functions_in_steps:
                if function not in child_functions:
                    raise mlrun.errors.MLRunInvalidArgumentError(
                        f"function {function} is used in steps and is not defined, "
                        "use the .add_child_function() to specify child function attributes"
                    )
        if (
            isinstance(self.spec.graph, RootFlowStep)
            and any(
                isinstance(step_type, mlrun.serving.states.ModelRunnerStep)
                for step_type in self.spec.graph.steps.values()
            )
            and self.spec.build.functionSourceCode
        ):
            # Add import for LLModel
            decoded_code = b64decode(self.spec.build.functionSourceCode).decode("utf-8")
            import_llmodel_code = "\nfrom mlrun.serving.states import LLModel\n"
            if import_llmodel_code not in decoded_code:
                decoded_code += import_llmodel_code
            encoded_code = mlrun.utils.helpers.encode_user_code(decoded_code)
            self.spec.build.functionSourceCode = encoded_code

        # Handle secret processing before handling child functions, since secrets are transferred to them
        if self.spec.secret_sources:
            # Before passing to remote builder, secrets values must be retrieved (for example from ENV)
            # and stored as inline secrets. Otherwise, they will not be available to the builder.
            self._secrets = SecretsStore.from_list(self.spec.secret_sources)
            self.spec.secret_sources = self._secrets.to_serial()

        if self._spec.function_refs:
            # ensure the function is available to the UI while deploying the child functions
            self.save(versioned=False)

            # deploy child functions
            self._add_ref_triggers()
            self._deploy_function_refs()
            logger.info(f"deploy root function {self.metadata.name} ...")

        return super().deploy(
            project,
            tag,
            verbose,
            builder_env=builder_env,
            force_build=force_build,
        )

    def _get_serving_spec(self):
        function_name_uri_map = {f.name: f.uri(self) for f in self.spec.function_refs}
        serving_spec = {
            "function_name": self.metadata.name,
            "function_tag": self.metadata.tag,
            "function_uri": self._function_uri(),
            "function_hash": self.metadata.hash,
            "project": self.metadata.project,
            "version": "v2",
            "parameters": self.spec.parameters,
            "graph": self.spec.graph.to_dict(strip=True) if self.spec.graph else {},
            "load_mode": self.spec.load_mode,
            "functions": function_name_uri_map,
            "graph_initializer": self.spec.graph_initializer,
            "error_stream": self.spec.error_stream,
            "track_models": self.spec.track_models,
            "default_content_type": self.spec.default_content_type,
            "model_endpoint_creation_task_name": self.spec.model_endpoint_creation_task_name,
            # TODO: find another way to pass this (needed for local run)
            "filename": getattr(self.spec, "filename", None),
        }

        if self.spec.secret_sources:
            self._secrets = SecretsStore.from_list(self.spec.secret_sources)
            serving_spec["secret_sources"] = self._secrets.to_serial()

        return json.dumps(serving_spec)

    @property
    def serving_spec(self):
        return self._get_serving_spec()

    def to_mock_server(
        self,
        namespace=None,
        current_function="*",
        track_models=False,
        workdir=None,
        stream_profile: Optional[ds_profile.DatastoreProfile] = None,
        **kwargs,
    ) -> GraphServer:
        """create mock server object for local testing/emulation

        :param namespace: one or list of namespaces/modules to search the steps classes/functions in
        :param current_function: specify if you want to simulate a child function, * for all functions
        :param track_models: allow model tracking (disabled by default in the mock server)
        :param workdir:   working directory to locate the source code (if not the current one)
        :param stream_profile:   stream profile to use for the mock server output stream.
        """

        # set the namespaces/modules to look for the steps code in
        namespace = namespace or []
        if not isinstance(namespace, list):
            namespace = [namespace]
        module = mlrun.run.function_to_module(self, silent=True, workdir=workdir)
        if module:
            namespace.append(module)
        namespace.append(get_caller_globals())

        if workdir:
            old_workdir = os.getcwd()
            workdir = os.path.realpath(workdir)
            set_paths(workdir)
            os.chdir(workdir)

        server = create_graph_server(
            parameters=self.spec.parameters,
            load_mode=self.spec.load_mode,
            graph=self.spec.graph,
            verbose=self.verbose,
            current_function=current_function,
            graph_initializer=self.spec.graph_initializer,
            track_models=self.spec.track_models,
            function_uri=self._function_uri(),
            secret_sources=self.spec.secret_sources,
            default_content_type=self.spec.default_content_type,
            function_name=self.metadata.name,
            function_tag=self.metadata.tag,
            project=self.metadata.project,
            **kwargs,
        )
        server.init_states(
            context=None,
            namespace=namespace,
            logger=logger,
            is_mock=True,
            monitoring_mock=self.spec.track_models,
            stream_profile=stream_profile,
        )

        server.graph = add_system_steps_to_graph(
            server.project,
            server.graph,
            self.spec.track_models,
            server.context,
            self.spec,
        )

        if workdir:
            os.chdir(old_workdir)

        server.init_object(namespace)
        return server

    def plot(self, filename=None, format=None, source=None, **kw):
        """plot/save graph using graphviz

        example::

            serving_fn = mlrun.new_function("serving", image="mlrun/mlrun", kind="serving")
            serving_fn.add_model(
                "my-classifier",
                model_path=model_path,
                class_name="mlrun.frameworks.sklearn.SKLearnModelServer",
            )
            serving_fn.plot(rankdir="LR")

        :param filename:  target filepath for the image (None for the notebook)
        :param format:    The output format used for rendering (``'pdf'``, ``'png'``, etc.)
        :param source:    source step to add to the graph
        :param kw:        kwargs passed to graphviz, e.g. rankdir="LR" (see: https://graphviz.org/doc/info/attrs.html)
        :return: graphviz graph object
        """
        return self.spec.graph.plot(filename, format=format, source=source, **kw)

    def _set_as_mock(self, enable):
        if not enable:
            self._mock_server = None
            return

        logger.info(
            "Deploying serving function MOCK (for simulation)...\n"
            "Turn off the mock (mock=False) and make sure Nuclio is installed for real deployment to Nuclio"
        )
        self._mock_server = self.to_mock_server()

    def to_job(self, func_name: Optional[str] = None) -> KubejobRuntime:
        """Convert this ServingRuntime to a KubejobRuntime, so that the graph can be run as a standalone job.

        Args:
            func_name: Optional custom name for the job function. If not provided, automatically
                      appends '-batch' suffix to the serving function name to prevent database collision.

        Returns:
            KubejobRuntime configured to execute the serving graph as a batch job.

        Note:
            The job will have a different name than the serving function to prevent database collision.
            The original serving function remains unchanged and can still be invoked after running the job.
        """
        if self.spec.function_refs:
            raise mlrun.errors.MLRunInvalidArgumentError(
                f"Cannot convert function '{self.metadata.name}' to a job because it has child functions"
            )

        spec = KubeResourceSpec(
            image=self.spec.image,
            mode=self.spec.mode,
            volumes=self.spec.volumes,
            volume_mounts=self.spec.volume_mounts,
            env=self.spec.env,
            resources=self.spec.resources,
            default_handler="mlrun.serving.server.execute_graph",
            pythonpath=self.spec.pythonpath,
            entry_points=self.spec.entry_points,
            description=self.spec.description,
            workdir=self.spec.workdir,
            image_pull_secret=self.spec.image_pull_secret,
            build=self.spec.build,
            node_name=self.spec.node_name,
            node_selector=self.spec.node_selector,
            affinity=self.spec.affinity,
            disable_auto_mount=self.spec.disable_auto_mount,
            priority_class_name=self.spec.priority_class_name,
            tolerations=self.spec.tolerations,
            preemption_mode=self.spec.preemption_mode,
            security_context=self.spec.security_context,
            state_thresholds=self.spec.state_thresholds,
            serving_spec=self._get_serving_spec(),
            track_models=self.spec.track_models,
            parameters=self.spec.parameters,
            graph=self.spec.graph,
        )

        job_metadata = deepcopy(self.metadata)
        original_name = job_metadata.name

        if func_name:
            # User provided explicit job name
            job_metadata.name = func_name
            logger.debug(
                "Creating job from serving function with custom name",
                new_name=func_name,
            )
        else:
            job_metadata.name, was_renamed, suffix = (
                mlrun.utils.helpers.ensure_batch_job_suffix(job_metadata.name)
            )

            # Check if the resulting name exceeds Kubernetes length limit
            if (
                len(job_metadata.name)
                > mlrun.common.constants.K8S_DNS_1123_LABEL_MAX_LENGTH
            ):
                raise mlrun.errors.MLRunInvalidArgumentError(
                    f"Cannot convert serving function '{original_name}' to batch job: "
                    f"the resulting name '{job_metadata.name}' ({len(job_metadata.name)} characters) "
                    f"exceeds Kubernetes limit of {mlrun.common.constants.K8S_DNS_1123_LABEL_MAX_LENGTH} characters. "
                    f"Please provide a custom name via the func_name parameter, "
                    f"with at most {mlrun.common.constants.K8S_DNS_1123_LABEL_MAX_LENGTH} characters."
                )

            if was_renamed:
                logger.info(
                    "Creating job from serving function (auto-appended suffix to prevent collision)",
                    new_name=job_metadata.name,
                    suffix=suffix,
                )
            else:
                logger.debug(
                    "Creating job from serving function (name already has suffix)",
                    name=original_name,
                    suffix=suffix,
                )

        job = KubejobRuntime(
            spec=spec,
            metadata=job_metadata,
        )
        return job



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Callable, Optional

import numpy as np
import onnx
import onnxruntime

import mlrun

from .dataset import ONNXDataset


# TODO: Finish evaluation and prediction.
class ONNXMLRunInterface:
    """
    An interface for enabling convenient MLRun features for the ONNX framework.
    """

    # MLRun's context default name:
    DEFAULT_CONTEXT_NAME = "mlrun-onnx"

    def __init__(
        self,
        model: onnx.ModelProto,
        execution_providers: Optional[list[str]] = None,
        context: mlrun.MLClientCtx = None,
    ):
        # Set the context:
        self._context = (
            context
            if context is not None
            else mlrun.get_or_create_ctx(self.DEFAULT_CONTEXT_NAME)
        )

        # Store the model:
        self._model = model

        # Set the execution providers (default will prefer CUDA Execution Provider over CPU Execution Provider):
        self._execution_providers = (
            ["CUDAExecutionProvider", "CPUExecutionProvider"]
            if execution_providers is None
            else execution_providers
        )

        # initialize the onnx run time session:
        self._inference_session = onnxruntime.InferenceSession(
            onnx._serialize(model),
            providers=self._execution_providers,
        )

        # Get the input layers names:
        self._input_layers = [
            input_layer.name for input_layer in self._inference_session.get_inputs()
        ]

        # Get the outputs layers names:
        self._output_layers = [
            output_layer.name for output_layer in self._inference_session.get_outputs()
        ]

    def evaluate(
        self,
        dataset: ONNXDataset,
        metrics: list[Callable[[np.ndarray, np.ndarray], float]],
    ):
        pass

    def predict(self, inputs: np.ndarray):
        pass



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import enum

import mlrun.common.schemas.notification as notifications
import mlrun.utils.notifications.notification.base as base
import mlrun.utils.notifications.notification.console as console
import mlrun.utils.notifications.notification.git as git
import mlrun.utils.notifications.notification.ipython as ipython
import mlrun.utils.notifications.notification.mail as mail
import mlrun.utils.notifications.notification.slack as slack
import mlrun.utils.notifications.notification.webhook as webhook


class NotificationTypes(str, enum.Enum):
    console = notifications.NotificationKind.console.value
    git = notifications.NotificationKind.git.value
    ipython = notifications.NotificationKind.ipython.value
    slack = notifications.NotificationKind.slack.value
    mail = notifications.NotificationKind.mail.value
    webhook = notifications.NotificationKind.webhook.value

    def get_notification(self) -> type[base.NotificationBase]:
        return {
            self.console: console.ConsoleNotification,
            self.git: git.GitNotification,
            self.ipython: ipython.IPythonNotification,
            self.slack: slack.SlackNotification,
            self.mail: mail.MailNotification,
            self.webhook: webhook.WebhookNotification,
        }.get(self)

    def inverse_dependencies(self) -> list[str]:
        """
        Some notifications should only run if another notification type didn't run.
        Per given notification type, return a list of notification types that should not run in order for this
        notification to run.
        """
        return {
            self.console: [self.ipython],
        }.get(self, [])

    @classmethod
    def local(cls) -> list[str]:
        return [
            cls.console,
            cls.ipython,
        ]

    @classmethod
    def all(cls) -> list[str]:
        return [
            cls.console,
            cls.git,
            cls.ipython,
            cls.slack,
            cls.mail,
            cls.webhook,
        ]



# Copyright 2025 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import enum

from pydantic.v1 import BaseModel

from mlrun.common.types import StrEnum

from .background_task import BackgroundTaskList


class DeployResponse(BaseModel):
    data: dict
    background_tasks: BackgroundTaskList


class ModelRunnerStepData(StrEnum):
    MODELS = "models"
    MODEL_TO_EXECUTION_MECHANISM = "execution_mechanism_by_model_name"
    MONITORING_DATA = "monitoring_data"


class MonitoringData(StrEnum):
    INPUTS = "inputs"
    OUTPUTS = "outputs"
    INPUT_PATH = "input_path"
    RESULT_PATH = "result_path"
    CREATION_STRATEGY = "creation_strategy"
    LABELS = "labels"
    MODEL_PATH = "model_path"
    MODEL_ENDPOINT_UID = "model_endpoint_uid"
    MODEL_CLASS = "model_class"


class ModelsData(enum.Enum):
    MODEL_CLASS = 0
    MODEL_PARAMETERS = 1


MAX_BATCH_JOB_DURATION = "1w"



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional

from ..._ml_common.loggers import Logger
from ..utils import LGBMTypes
from .callback import Callback, CallbackEnv


class LoggingCallback(Callback):
    """
    A logging callback to collect training data.
    """

    def __init__(
        self,
        dynamic_hyperparameters: Optional[list[str]] = None,
        static_hyperparameters: Optional[list[str]] = None,
    ):
        """
        Initialize the logging callback with the given configuration. All the metrics data will be collected but the
        hyperparameters to log must be given. The hyperparameters will be taken from the `params` of the model in each
        iteration.

        :param dynamic_hyperparameters: If needed to track a hyperparameter dynamically (sample it each iteration) it
                                        should be passed here. The parameter expects a list of all the hyperparameters
                                        names to track our of the `params` dictionary.
        :param static_hyperparameters:  If needed to track a hyperparameter one time per run it should be passed here.
                                        The parameter expects a list of all the hyperparameters names to track our of
                                        the `params` dictionary.
        """
        super().__init__()
        self._logger = Logger()
        self._dynamic_hyperparameters_keys = (
            dynamic_hyperparameters if dynamic_hyperparameters is not None else {}
        )
        self._static_hyperparameters_keys = (
            static_hyperparameters if static_hyperparameters is not None else {}
        )

    @property
    def logger(self) -> Logger:
        """
        Get the logger of the callback. In the logger you may access the collected training data.

        :return: The logger.
        """
        return self._logger

    def __call__(self, env: CallbackEnv):
        """
        Log the iteration that ended and all the results it calculated.

        :param env: A named tuple passed ad the end of each iteration containing the metrics results. For more
                    information check the `Callback` doc string.
        """
        # Log the iteration:
        self._logger.log_iteration()

        # Log the metrics results out of the `evaluation_result_list` field:
        self._log_results(evaluation_result_list=env.evaluation_result_list)

        # Log the hyperparameters out of the `params` field:
        self._log_hyperparameters(parameters=env.params)

    def _log_results(
        self, evaluation_result_list: list[LGBMTypes.EvaluationResultType]
    ):
        """
        Log the callback environment results data into the logger.

        :param evaluation_result_list: The metrics results as provided by the callback environment of LightGBM.
        """
        for evaluation_result in evaluation_result_list:
            # Check what results were given, from `lightgbm.train` or `lightgbm.cv`:
            if len(evaluation_result) == 4:
                # `lightgbm.train` is used:
                self._logger.log_result(
                    validation_set_name=evaluation_result[0],
                    metric_name=evaluation_result[1],
                    result=evaluation_result[2],
                )
            else:
                # `lightgbm.cv` is used, unpack both mean and stdv scores:
                self._logger.log_result(
                    validation_set_name=evaluation_result[0],
                    metric_name=f"{evaluation_result[1]}_mean",
                    result=evaluation_result[2],
                )
                self._logger.log_result(
                    validation_set_name=evaluation_result[0],
                    metric_name=f"{evaluation_result[1]}_stdv",
                    result=evaluation_result[4],
                )

    def _log_hyperparameters(self, parameters: dict):
        """
        Log the callback environment parameters into the logger.

        :param parameters: The parameters as provided by the callback environment of LightGBM.
        """
        for parameter_name, value in parameters.items():
            if parameter_name in self._dynamic_hyperparameters_keys:
                self._logger.log_dynamic_hyperparameter(
                    parameter_name=parameter_name, value=value
                )
                continue
            if parameter_name in self._static_hyperparameters_keys:
                self._logger.log_static_hyperparameter(
                    parameter_name=parameter_name, value=value
                )



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from dependency_injector import containers, providers

import mlrun.config
import mlrun.errors
import mlrun.launcher.base
import mlrun.launcher.local
import mlrun.launcher.remote
import mlrun.utils.singleton


class LauncherFactory(
    metaclass=mlrun.utils.singleton.AbstractSingleton,
):
    def __init__(self):
        self._launcher_container = LauncherContainer()

    def create_launcher(
        self, is_remote: bool, **kwargs
    ) -> mlrun.launcher.base.BaseLauncher:
        """
        Creates the appropriate launcher for the specified run.
        ServerSideLauncher - if running as API.
        ClientRemoteLauncher - if the run is remote and local was not specified.
        ClientLocalLauncher - if the run is not remote or local was specified.

        :param is_remote:   Whether the runtime requires remote execution.

        :return:            The appropriate launcher for the specified run.
        """
        if mlrun.config.is_running_as_api():
            return self._launcher_container.server_side_launcher(**kwargs)

        local = kwargs.get("local", False)
        if is_remote and not local:
            return self._launcher_container.client_remote_launcher(**kwargs)

        return self._launcher_container.client_local_launcher(**kwargs)


class LauncherContainer(containers.DeclarativeContainer):
    client_remote_launcher = providers.Factory(
        mlrun.launcher.remote.ClientRemoteLauncher
    )
    client_local_launcher = providers.Factory(mlrun.launcher.local.ClientLocalLauncher)

    # Provider for injection of a server side launcher.
    # This allows us to override the launcher from external packages without having to import them.
    server_side_launcher = providers.Factory(mlrun.launcher.base.BaseLauncher)
